#!/usr/bin/env python3
"""ClawPolicy command-line interface."""

from __future__ import annotations

import argparse
import json
import shutil
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

from .api import ConfirmationAPI
from .confirmation import IntelligentConfirmation
from .md_to_policy import MarkdownToPolicyConverter
from .policy_store import PolicyStore
from .policy_to_md import PolicyToMarkdownExporter


class ClawPolicyCLI:
    """Main ClawPolicy CLI class."""

    def __init__(self) -> None:
        self.memory_dir_name = ".clawpolicy"
        self.config_file_name = "config.json"
        self.templates = {
            "USER.md": "USER_template.md",
            "SOUL.md": "SOUL_template.md",
            "AGENTS.md": "AGENTS_template.md",
        }

    def get_template_dir(self) -> Path:
        package_dir = Path(__file__).parent
        return package_dir.parent / "templates"

    def get_memory_dir(self, cwd: Optional[Path] = None) -> Path:
        return (cwd or Path.cwd()) / self.memory_dir_name

    def get_policy_dir(self, cwd: Optional[Path] = None) -> Path:
        return self.get_memory_dir(cwd) / "policy"

    def _open_policy_store(
        self,
        cwd: Optional[Path] = None,
        *,
        ensure_files: bool = False,
    ) -> PolicyStore:
        return PolicyStore.bootstrap(self.get_memory_dir(cwd), ensure_files=ensure_files)

    def init(self, target_dir: Optional[str] = None, force: bool = False) -> bool:
        cwd = Path(target_dir).resolve() if target_dir else Path.cwd()
        memory_dir = self.get_memory_dir(cwd)
        template_dir = self.get_template_dir()

        if not template_dir.exists():
            print(f"Error: template directory not found: {template_dir}")
            return False

        memory_dir.mkdir(parents=True, exist_ok=True)
        store = self._open_policy_store(cwd)
        policy_dir = store.base_dir

        if not force and self._has_markdown_seed(memory_dir) and not store.load_rules() and not store.load_playbooks():
            MarkdownToPolicyConverter().migrate_all(memory_dir, store)

        if force or not store.rules_file.exists():
            store.save_rules({})
            print(f"Created: {store.rules_file}")
        if force or not store.playbooks_file.exists():
            store.save_playbooks({})
            print(f"Created: {store.playbooks_file}")
        if force or not store.policy_events_file.exists():
            store.policy_events_file.touch()
            print(f"Created: {store.policy_events_file}")

        for target_name, template_name in self.templates.items():
            template_file = template_dir / template_name
            target_file = memory_dir / target_name
            if not template_file.exists():
                continue
            if target_file.exists() and not force:
                continue
            shutil.copy2(template_file, target_file)
            print(f"Created: {target_file}")

        config_file = memory_dir / self.config_file_name
        if force or not config_file.exists():
            config = {
                "version": "3.0.0",
                "initialized_at": datetime.now(timezone.utc).isoformat(),
                "initialized_in": str(cwd),
                "memory_path": str(memory_dir),
                "features": {
                    "auto_learning": True,
                    "safety_checks": True,
                    "policy_engine": "autonomous_execution",
                },
            }
            config_file.write_text(json.dumps(config, indent=2, ensure_ascii=False), encoding="utf-8")
            print(f"Created: {config_file}")

        gitignore_file = memory_dir / ".gitignore"
        if force or not gitignore_file.exists():
            gitignore_file.write_text(
                "# ClawPolicy local files\n"
                "# Do not commit these files\n"
                "config.json\n"
                "*.backup\n"
                "*.cache\n",
                encoding="utf-8",
            )
            print(f"Created: {gitignore_file}")

        print("")
        print("Initialization complete.")
        print(f"Policy memory: {memory_dir}")
        print(f"Policy assets: {policy_dir}")
        return True

    def _has_markdown_seed(self, memory_dir: Path) -> bool:
        return any((memory_dir / name).exists() for name in self.templates)

    def status(self) -> None:
        memory_dir = self.get_memory_dir()
        store = self._open_policy_store()
        policy_dir = store.base_dir
        config_file = memory_dir / self.config_file_name

        print("ClawPolicy status")
        print("")
        print(f"Policy memory: {memory_dir}")
        print(f"Status: {'present' if memory_dir.exists() else 'missing'}")
        print("")

        if config_file.exists():
            config = json.loads(config_file.read_text(encoding="utf-8"))
            print(f"Config version: {config.get('version', 'unknown')}")
        else:
            print("Config version: missing")

        stats = store.get_stats()
        print(f"Rules: {stats['total_rules']}")
        print(f"Playbooks: {stats['total_playbooks']}")
        print(f"Policy events: {stats['total_policy_events']}")
        print(f"Policy directory: {policy_dir}")

    def version(self) -> None:
        from . import __version__

        print(f"ClawPolicy CLI v{__version__}")

    def rule_list(self) -> None:
        store = self._open_policy_store()
        rules = store.load_rules()
        if not rules:
            print("No rules found.")
            return
        print(f"Rules ({len(rules)} total):")
        for rule in rules.values():
            print(f"  {rule}")

    def rule_show(self, rule_id: str) -> None:
        store = self._open_policy_store()
        rule = store.get_rule(rule_id)
        if rule is None:
            print(f"Rule not found: {rule_id}")
            return

        print(f"Rule Details: {rule.id}")
        print(f"Summary: {rule.summary}")
        print(f"Status: {rule.status}")
        print(f"Scope: {rule.scope}")
        if rule.scope_key:
            print(f"Scope Key: {rule.scope_key}")
        print(f"Policy Decision: {rule.policy_decision or '(none)'}")
        print(f"Evidence Count: {rule.evidence_count}")
        print(f"Risk Level: {rule.risk_level or '(unset)'}")
        if rule.validation:
            print("Validation:")
            for test in rule.validation:
                print(f"  - {test}")

    def playbook_list(self) -> None:
        store = self._open_policy_store()
        playbooks = store.load_playbooks()
        if not playbooks:
            print("No playbooks found.")
            return
        print(f"Playbooks ({len(playbooks)} total):")
        for playbook in playbooks.values():
            print(f"  {playbook}")

    def playbook_show(self, playbook_id: str) -> None:
        store = self._open_policy_store()
        playbook = store.get_playbook(playbook_id)
        if playbook is None:
            print(f"Playbook not found: {playbook_id}")
            return
        print(f"Playbook Details: {playbook.id}")
        print(f"Summary: {playbook.summary}")
        print(f"Category: {playbook.category}")
        print(f"Confidence: {playbook.confidence:.2f}")
        print(f"Rules Used: {', '.join(playbook.rules_used) if playbook.rules_used else '(none)'}")

    def events_show(self, limit: int = 20) -> None:
        store = self._open_policy_store()
        events = store.get_events(limit)
        if not events:
            print("No policy events found.")
            return
        print(f"Recent policy events ({len(events)} shown):")
        for event in events:
            print(f"  {event}")

    def export_md(self) -> None:
        memory_dir = self.get_memory_dir()
        PolicyToMarkdownExporter().export_all(self._open_policy_store().base_dir, memory_dir)
        print("Markdown export completed.")

    def confidence_history(self, task_type: Optional[str] = None) -> None:
        rules = ConfirmationAPI(memory_dir=self.get_memory_dir()).get_confidence_history(task_type)["rules"]
        if not rules:
            print("No rules found.")
            return
        print(f"Confidence History ({len(rules)} rules)")
        print("")
        print(f"{'Confidence':<12} {'Streak':<8} {'Status':<12} {'Scope':<10} {'Rule Summary'}")
        print("-" * 88)
        for rule in rules:
            print(
                f"{rule['confidence']:<12.2f} {rule['success_streak']:<8} "
                f"{rule['status']:<12} {rule['scope']:<10} {rule['summary']}"
            )

    def decision_history(self, limit: int = 10) -> None:
        decisions = ConfirmationAPI(memory_dir=self.get_memory_dir()).get_recent_decisions(limit)["decisions"]
        if not decisions:
            print("No decision history found.")
            return
        print(f"Recent Decisions ({len(decisions)} shown)")
        print("")
        for decision in decisions:
            outcome = decision.get("execution_result") or "pending"
            print(
                f"- {decision['decision_id']} | {decision['final_decision']} | "
                f"risk={decision['risk_level']} | outcome={outcome}"
            )
            print(f"  summary: {decision['task_summary']}")
            print(f"  reason: {decision['reason']}")
            print(f"  resolution: {decision['resolution']}")

    def policy_status(self) -> None:
        snapshot = self._open_policy_store().get_policy_status_snapshot()
        print("Policy lifecycle status")
        print("")
        for status, count in snapshot["status_counts"].items():
            print(f"  {status}: {count}")
        print("")
        print(f"Recent promotions: {len(snapshot['recent_promotions'])}")
        print(f"Recent suspensions: {len(snapshot['recent_suspensions'])}")
        print(f"High-risk confirmed: {len(snapshot['risky_confirmed_rules'])}")
        print(f"Special scopes: {len(snapshot['special_scopes'])}")

    def policy_recent(self, limit: int = 10) -> None:
        events = self._open_policy_store().get_recent_lifecycle_events(limit=limit)
        if not events:
            print("No lifecycle events found.")
            return
        print(f"Recent policy lifecycle events ({len(events)} shown)")
        for event in events:
            payload = event.payload
            print(
                f"- {event.event_type} | {payload.get('rule_id', 'unknown')} | "
                f"{payload.get('trigger', 'n/a')}"
            )

    def policy_risky(self) -> None:
        rules = self._open_policy_store().get_risky_confirmed_rules()
        if not rules:
            print("No high-risk confirmed auto-execute rules.")
            return
        print("High-risk confirmed auto-execute rules")
        for rule in rules:
            print(f"- {rule.id} | {rule.risk_level} | {rule.scope}:{rule.scope_key or '*'}")
            print(f"  {rule.summary}")

    def policy_suspended(self, limit: int = 20) -> None:
        rules = self._open_policy_store().get_rules_by_status("suspended")[:limit]
        if not rules:
            print("No suspended rules.")
            return
        print(f"Suspended rules ({len(rules)} shown)")
        for rule in rules:
            print(
                f"- {rule.id} | {rule.scope}:{rule.scope_key or '*'} | "
                f"{rule.suspension_reason or 'unknown'}"
            )
            print(f"  {rule.summary}")

    def execute_demo(self, task_type: str = "T2", description: str = "run tests") -> None:
        store = self._open_policy_store()
        engine = IntelligentConfirmation(store)
        task_context = {
            "task_type": task_type,
            "task_description": description,
            "command": f"npm run {description}",
            "files": [],
        }
        should_confirm, reason = engine.should_confirm(task_context)
        print(engine.get_explanation(task_context, should_confirm, reason))


def main() -> None:
    parser = argparse.ArgumentParser(
        description="ClawPolicy - explainable autonomous execution policy engine",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("--version", action="store_true", help="Show version information")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    init_parser = subparsers.add_parser("init", help="Initialize memory files")
    init_parser.add_argument("target_dir", nargs="?", help="Target directory")
    init_parser.add_argument("--force", action="store_true", help="Force overwrite existing files")

    subparsers.add_parser("status", help="Show current status")

    analyze_parser = subparsers.add_parser("analyze", help="Analyze Git history and learn preferences")
    analyze_parser.add_argument("--repo", default=".", help="Git repository path")
    analyze_parser.add_argument("--commits", type=int, default=100, help="Number of commits to analyze")

    rule_parser = subparsers.add_parser("rule", help="Rule management commands")
    rule_subparsers = rule_parser.add_subparsers(dest="rule_command", help="Rule actions")
    rule_subparsers.add_parser("list", help="List all rules")
    rule_show_parser = rule_subparsers.add_parser("show", help="Show rule details")
    rule_show_parser.add_argument("rule_id", help="Rule ID")

    playbook_parser = subparsers.add_parser("playbook", help="Playbook management commands")
    playbook_subparsers = playbook_parser.add_subparsers(dest="playbook_command", help="Playbook actions")
    playbook_subparsers.add_parser("list", help="List all playbooks")
    playbook_show_parser = playbook_subparsers.add_parser("show", help="Show playbook details")
    playbook_show_parser.add_argument("playbook_id", help="Playbook ID")

    events_parser = subparsers.add_parser("events", help="Show policy events")
    events_parser.add_argument("--limit", type=int, default=20, help="Number of events to show")

    subparsers.add_parser("export-md", help="Export policy memory to Markdown format")

    confidence_parser = subparsers.add_parser("confidence-history", help="Show confidence history")
    confidence_parser.add_argument("--task-type", help="Filter by task type")

    decision_parser = subparsers.add_parser("decision-history", help="Show recent confirmation decisions")
    decision_parser.add_argument("--limit", type=int, default=10, help="Number of recent decisions to show")

    policy_parser = subparsers.add_parser("policy", help="Read-only policy lifecycle operations")
    policy_subparsers = policy_parser.add_subparsers(dest="policy_command", help="Policy actions")
    policy_subparsers.add_parser("status", help="Show policy lifecycle status")
    policy_recent_parser = policy_subparsers.add_parser("recent", help="Show recent lifecycle events")
    policy_recent_parser.add_argument("--limit", type=int, default=10, help="Number of events to show")
    policy_subparsers.add_parser("risky", help="Show risky confirmed rules")
    policy_suspended_parser = policy_subparsers.add_parser("suspended", help="Show suspended rules")
    policy_suspended_parser.add_argument("--limit", type=int, default=20, help="Number of rules to show")

    execute_parser = subparsers.add_parser("execute-demo", help="Demonstrate confirmation workflow")
    execute_parser.add_argument("--task-type", default="T2", help="Task type")
    execute_parser.add_argument("--description", default="run tests", help="Task description")

    args = parser.parse_args()
    cli = ClawPolicyCLI()

    if args.version:
        cli.version()
        return
    if args.command == "init":
        raise SystemExit(0 if cli.init(args.target_dir, args.force) else 1)
    if args.command == "status":
        cli.status()
        return
    if args.command == "analyze":
        from .integration import IntentAlignmentEngine

        IntentAlignmentEngine(args.repo).run_analysis(args.commits)
        return
    if args.command == "rule":
        if args.rule_command == "list":
            cli.rule_list()
        elif args.rule_command == "show":
            cli.rule_show(args.rule_id)
        else:
            rule_parser.print_help()
        return
    if args.command == "playbook":
        if args.playbook_command == "list":
            cli.playbook_list()
        elif args.playbook_command == "show":
            cli.playbook_show(args.playbook_id)
        else:
            playbook_parser.print_help()
        return
    if args.command == "events":
        cli.events_show(args.limit)
        return
    if args.command == "export-md":
        cli.export_md()
        return
    if args.command == "confidence-history":
        cli.confidence_history(args.task_type)
        return
    if args.command == "decision-history":
        cli.decision_history(args.limit)
        return
    if args.command == "policy":
        if args.policy_command == "status":
            cli.policy_status()
        elif args.policy_command == "recent":
            cli.policy_recent(args.limit)
        elif args.policy_command == "risky":
            cli.policy_risky()
        elif args.policy_command == "suspended":
            cli.policy_suspended(args.limit)
        else:
            policy_parser.print_help()
        return
    if args.command == "execute-demo":
        cli.execute_demo(args.task_type, args.description)
        return

    parser.print_help()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Focused tests for policy promotion and scope precedence."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from pathlib import Path

from lib.api import ConfirmationAPI
from lib.policy_models import Rule
from lib.policy_store import PolicyStore


def _iso(minutes: int = 0) -> str:
    return (datetime.now(timezone.utc) + timedelta(minutes=minutes)).isoformat()


def test_explicit_correction_creates_candidate_rule(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
    }

    _, _, details = api.should_auto_execute(task)
    api.record_execution_result(
        task=task,
        success=True,
        auto_executed=False,
        decision_id=details["decision_id"],
        execution_result="success",
        user_override="confirmed_after_prompt",
    )

    rules = PolicyStore(memory_dir / "policy").load_rules()
    candidate_rules = [rule for rule in rules.values() if rule.status == "candidate"]

    assert candidate_rules
    assert candidate_rules[0].source_type in {"explicit_correction", "explicit_preference"}


def test_candidate_promotes_after_gate_when_evidence_is_strong(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
    }

    for _ in range(3):
        _, _, details = api.should_auto_execute(task)
        api.record_execution_result(
            task=task,
            success=True,
            auto_executed=False,
            decision_id=details["decision_id"],
            execution_result="success",
            user_override="confirmed_after_prompt",
        )

    rules = PolicyStore(memory_dir / "policy").load_rules()
    confirmed_rules = [rule for rule in rules.values() if rule.status == "confirmed"]

    assert confirmed_rules
    assert all(rule.validation for rule in confirmed_rules)


def test_same_scope_newer_confirmed_rule_wins(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    project_scope = str(tmp_path.resolve())
    store = PolicyStore(memory_dir / "policy")

    older_rule = Rule(
        id="older_rule",
        summary="Older project rule says confirm",
        category="harden",
        trigger=["task_type:T2", "command:python -m pytest tests/"],
        status="confirmed",
        scope="project",
        scope_key=project_scope,
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at=_iso(-10),
        policy_decision="require_confirmation",
        validation=["assert confirmation for pytest"],
    )
    older_rule.calculate_asset_id()

    newer_rule = Rule(
        id="newer_rule",
        summary="Newer project rule says auto",
        category="harden",
        trigger=["task_type:T2", "command:python -m pytest tests/"],
        status="confirmed",
        scope="project",
        scope_key=project_scope,
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at=_iso(0),
        policy_decision="auto_execute",
        validation=["assert auto execute for pytest"],
    )
    newer_rule.calculate_asset_id()

    store.save_rule(older_rule)
    store.save_rule(newer_rule)

    auto_execute, _, details = api.should_auto_execute(
        {
            "task_type": "T2",
            "task_description": "Run tests",
            "command": "python -m pytest tests/",
            "project_path": project_scope,
        }
    )

    assert auto_execute is True
    assert details["decision"]["resolution"] == "same_scope_newer_confirmed"


def test_high_risk_candidate_does_not_confirm_from_statistics_only(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    project_scope = str(tmp_path.resolve())
    store = PolicyStore(memory_dir / "policy")

    candidate = Rule(
        id="dangerous_auto_rule",
        summary="Auto-delete workspace",
        category="harden",
        trigger=["task_type:T3", "command:rm -rf workspace"],
        status="candidate",
        scope="project",
        scope_key=project_scope,
        evidence_count=2,
        source_type="repeated_success",
        last_seen_at=_iso(-5),
        policy_decision="auto_execute",
        validation=["assert auto execute for rm -rf workspace"],
        confidence=0.9,
        success_streak=5,
    )
    candidate.calculate_asset_id()
    store.save_rule(candidate)

    for _ in range(2):
        _, _, details = api.should_auto_execute(
            {
                "task_type": "T3",
                "task_description": "Clean workspace directory",
                "command": "rm -rf workspace",
                "project_path": project_scope,
            }
        )
        api.record_execution_result(
            task={
                "task_type": "T3",
                "task_description": "Clean workspace directory",
                "command": "rm -rf workspace",
                "project_path": project_scope,
            },
            success=True,
            auto_executed=True,
            decision_id=details["decision_id"],
            execution_result="success",
        )

    refreshed = store.get_rule("dangerous_auto_rule")
    assert refreshed is not None
    assert refreshed.status == "candidate"


def test_domain_confirmed_rule_overrides_global_when_project_rule_missing(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    store = PolicyStore(memory_dir / "policy")

    global_rule = Rule(
        id="global_docs_rule",
        summary="Docs changes require confirmation globally",
        category="harden",
        trigger=["task_type:T2", "command:python -m mkdocs build"],
        status="confirmed",
        scope="global",
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at=_iso(-10),
        policy_decision="require_confirmation",
        validation=["assert confirmation for docs build"],
    )
    global_rule.calculate_asset_id()

    domain_rule = Rule(
        id="domain_docs_rule",
        summary="Docs builds auto execute in docs domain",
        category="harden",
        trigger=["task_type:T2", "command:python -m mkdocs build"],
        status="confirmed",
        scope="domain",
        scope_key="docs",
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at=_iso(0),
        policy_decision="auto_execute",
        validation=["assert auto execute for docs build"],
    )
    domain_rule.calculate_asset_id()

    store.save_rule(global_rule)
    store.save_rule(domain_rule)

    auto_execute, _, details = api.should_auto_execute(
        {
            "task_type": "T2",
            "task_description": "Build docs site",
            "command": "python -m mkdocs build",
            "files": ["docs/index.md"],
        }
    )

    assert auto_execute is True
    assert details["decision"]["resolution"] == "domain_over_global"
    assert details["decision"]["scope"] == "domain"
    assert details["decision"]["scope_key"] == "docs"


def test_confirmed_rule_suspends_after_repeated_auto_execute_failures(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    project_scope = str(tmp_path.resolve())
    store = PolicyStore(memory_dir / "policy")

    confirmed = Rule(
        id="confirmed_pytest_rule",
        summary="Project pytest runs may auto execute",
        category="harden",
        trigger=["task_type:T2", "command:python -m pytest tests/"],
        status="confirmed",
        scope="project",
        scope_key=project_scope,
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at=_iso(-5),
        policy_decision="auto_execute",
        validation=["assert auto execute for pytest"],
        confidence=0.9,
    )
    confirmed.calculate_asset_id()
    store.save_rule(confirmed)

    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
        "project_path": project_scope,
    }

    for _ in range(2):
        _, _, details = api.should_auto_execute(task)
        api.record_execution_result(
            task=task,
            success=False,
            auto_executed=True,
            decision_id=details["decision_id"],
            execution_result="failure",
        )

    refreshed = store.get_rule("confirmed_pytest_rule")
    lifecycle_events = store.get_events_by_type({"rule_suspended"}, limit=10)

    assert refreshed is not None
    assert refreshed.status == "suspended"
    assert lifecycle_events
    assert lifecycle_events[0].payload["rule_id"] == "confirmed_pytest_rule"


def test_suspended_rule_reactivates_to_candidate_after_explicit_reapproval(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    project_scope = str(tmp_path.resolve())
    store = PolicyStore(memory_dir / "policy")

    suspended = Rule(
        id="suspended_pytest_rule",
        summary="Project pytest runs were suspended",
        category="harden",
        trigger=["task_type:T2", "command:python -m pytest tests/"],
        status="suspended",
        scope="project",
        scope_key=project_scope,
        evidence_count=5,
        source_type="explicit_preference",
        last_seen_at=_iso(-5),
        policy_decision="auto_execute",
        validation=["assert auto execute for pytest"],
        confidence=0.4,
    )
    suspended.calculate_asset_id()
    store.save_rule(suspended)

    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
        "project_path": project_scope,
    }

    for _ in range(2):
        _, _, details = api.should_auto_execute(task)
        api.record_execution_result(
            task=task,
            success=True,
            auto_executed=False,
            decision_id=details["decision_id"],
            execution_result="success",
            user_override="confirmed_after_prompt",
        )

    refreshed = store.get_rule("suspended_pytest_rule")
    lifecycle_events = store.get_events_by_type({"rule_reactivated"}, limit=10)

    assert refreshed is not None
    assert refreshed.status == "candidate"
    assert lifecycle_events
    assert lifecycle_events[0].payload["rule_id"] == "suspended_pytest_rule"


def test_rollback_immediately_suspends_confirmed_auto_rule(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    project_scope = str(tmp_path.resolve())
    store = PolicyStore(memory_dir / "policy")

    confirmed = Rule(
        id="rollback_sensitive_rule",
        summary="Project deploy preview may auto execute",
        category="harden",
        trigger=["task_type:T2", "command:deploy preview"],
        status="confirmed",
        scope="project",
        scope_key=project_scope,
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at=_iso(-5),
        policy_decision="auto_execute",
        validation=["assert auto execute for deploy preview"],
        confidence=0.9,
    )
    confirmed.calculate_asset_id()
    store.save_rule(confirmed)

    task = {
        "task_type": "T2",
        "task_description": "Deploy preview",
        "command": "deploy preview",
        "project_path": project_scope,
    }

    _, _, details = api.should_auto_execute(task)
    api.record_execution_result(
        task=task,
        success=False,
        auto_executed=True,
        decision_id=details["decision_id"],
        execution_result="rollback",
    )

    refreshed = store.get_rule("rollback_sensitive_rule")
    lifecycle_events = store.get_events_by_type({"rule_suspended"}, limit=10)

    assert refreshed is not None
    assert refreshed.status == "suspended"
    assert lifecycle_events[0].payload["trigger"] == "rollback"


def test_ambiguous_same_scope_conflict_requires_confirmation(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    project_scope = str(tmp_path.resolve())
    store = PolicyStore(memory_dir / "policy")

    first = Rule(
        id="conflict_rule_one",
        summary="Require confirmation for pytest",
        category="harden",
        trigger=["task_type:T2", "command:python -m pytest tests/"],
        status="confirmed",
        scope="project",
        scope_key=project_scope,
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at="2026-03-07T12:00:00+00:00",
        policy_decision="require_confirmation",
        validation=["assert confirmation for pytest"],
    )
    first.calculate_asset_id()

    second = Rule(
        id="conflict_rule_two",
        summary="Auto execute pytest",
        category="harden",
        trigger=["task_type:T2", "command:python -m pytest tests/"],
        status="confirmed",
        scope="project",
        scope_key=project_scope,
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at="2026-03-07T12:00:00+00:00",
        policy_decision="auto_execute",
        validation=["assert auto execute for pytest"],
    )
    second.calculate_asset_id()

    store.save_rule(first)
    store.save_rule(second)

    auto_execute, reason, details = api.should_auto_execute(
        {
            "task_type": "T2",
            "task_description": "Run tests",
            "command": "python -m pytest tests/",
            "project_path": project_scope,
        }
    )

    assert auto_execute is False
    assert "confirmation" in reason.lower()
    assert details["decision"]["resolution"] == "ambiguous_same_scope_conflict"


def test_suspended_rule_archives_when_superseded_by_new_confirmed_rule(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    project_scope = str(tmp_path.resolve())
    store = PolicyStore(memory_dir / "policy")

    suspended = Rule(
        id="superseded_suspended_rule",
        summary="Old suspended pytest policy",
        category="harden",
        trigger=["task_type:T2", "command:python -m pytest tests/"],
        status="suspended",
        scope="project",
        scope_key=project_scope,
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at=_iso(-10),
        policy_decision="require_confirmation",
        validation=["assert confirmation for pytest"],
    )
    suspended.calculate_asset_id()

    confirmed = Rule(
        id="confirmed_replacement_rule",
        summary="New confirmed pytest policy",
        category="harden",
        trigger=["task_type:T2", "command:python -m pytest tests/"],
        status="confirmed",
        scope="project",
        scope_key=project_scope,
        evidence_count=4,
        source_type="explicit_preference",
        last_seen_at=_iso(0),
        policy_decision="auto_execute",
        validation=["assert auto execute for pytest"],
    )
    confirmed.calculate_asset_id()

    store.save_rule(suspended)
    store.save_rule(confirmed)

    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
        "project_path": project_scope,
    }

    _, _, details = api.should_auto_execute(task)
    api.record_execution_result(
        task=task,
        success=True,
        auto_executed=True,
        decision_id=details["decision_id"],
        execution_result="success",
    )

    archived = store.get_rule("superseded_suspended_rule")
    lifecycle_events = store.get_events_by_type({"rule_archived"}, limit=10)

    assert archived is not None
    assert archived.status == "archived"
    assert lifecycle_events
    assert lifecycle_events[0].payload["rule_id"] == "superseded_suspended_rule"



#!/usr/bin/env python3
"""
data collection module - fromGitCollection of preference data from history and user actions
"""

import subprocess
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List


class GitPreferenceCollector:
    """fromGitHistory Collection Technology Preferences"""

    def __init__(self, repo_path: str = "."):
        self.repo_path = Path(repo_path).resolve()
        self.preferences: Dict[str, Any] = {
            "tech_stack": {},
            "file_types": {},
            "commit_patterns": {},
            "workflow": {}
        }

    def collect(self, max_commits: int = 100) -> Dict[str, Any]:
        """collectGitHistorical preference data"""
        print(f"📊 AnalyzingGithistory（recent{max_commits}commits）...")

        # Get commit history
        commits = self._get_commits(max_commits)
        print(f"✅ analyzed {len(commits)} commits")

        # Analysis Technology Stack
        self.preferences["tech_stack"] = self._analyze_tech_stack(commits)

        # Analyze file types
        self.preferences["file_types"] = self._analyze_file_types(commits)

        # Analytical workflow patterns
        self.preferences["workflow"] = self._analyze_workflow(commits)

        # Add metadata
        self.preferences["metadata"] = {
            "collected_at": datetime.now().isoformat(),
            "repo_path": str(self.repo_path),
            "commits_analyzed": len(commits),
            "confidence": self._calculate_confidence()
        }

        return self.preferences

    def _get_commits(self, max_count: int) -> List[Dict[str, Any]]:
        """GetGitCommit history"""
        try:
            result = subprocess.run(
                ["git", "log", f"-{max_count}", "--pretty=format:%H|%s|%an", "--name-only"],
                cwd=self.repo_path,
                capture_output=True,
                text=True,
                timeout=30
            )

            if result.returncode != 0:
                print("⚠️  Unable to obtainGithistory")
                return []

            commits: List[Dict[str, Any]] = []
            lines = result.stdout.strip().split('\n')

            current_commit: Dict[str, Any] | None = None
            for line in lines:
                if '|' in line:  # Submit information line
                    parts = line.split('|')
                    current_commit = {
                        "hash": parts[0],
                        "subject": parts[1],
                        "author": parts[2],
                        "files": []
                    }
                    commits.append(current_commit)
                elif current_commit and line:  # file line
                    current_commit["files"].append(line)

            return commits

        except subprocess.TimeoutExpired:
            print("⚠️  GitCommand timeout")
            return []
        except Exception as e:
            print(f"⚠️  GetGithistorical failure: {e}")
            return []

    def _analyze_tech_stack(self, commits: List[Dict[str, Any]]) -> Dict[str, int]:
        """Analyze technology stack preferences"""
        tech_stack = {
            "python": 0,
            "javascript": 0,
            "typescript": 0,
            "react": 0,
            "vue": 0,
            "fastapi": 0,
            "node": 0
        }

        for commit in commits:
            for file_path in commit.get("files", []):
                file_lower = file_path.lower()

                # DetectionPython
                if file_path.endswith('.py'):
                    tech_stack["python"] += 1

                # DetectionJavaScript/TypeScript
                if file_path.endswith('.js'):
                    tech_stack["javascript"] += 1
                if file_path.endswith('.ts') or file_path.endswith('.tsx'):
                    tech_stack["typescript"] += 1

                # DetectionReact
                if "react" in file_lower or file_path.endswith('.jsx'):
                    tech_stack["react"] += 1

                # DetectionVue
                if "vue" in file_lower:
                    tech_stack["vue"] += 1

                # DetectionFastAPI
                if "fastapi" in file_lower:
                    tech_stack["fastapi"] += 1

                # DetectionNode
                if "package.json" in file_path:
                    tech_stack["node"] += 1

        return tech_stack

    def _analyze_file_types(self, commits: List[Dict[str, Any]]) -> Dict[str, int]:
        """Analyze file type preferences"""
        file_types: Dict[str, int] = {}

        for commit in commits:
            for file_path in commit.get("files", []):
                ext = Path(file_path).suffix or "(No suffix)"
                file_types[ext] = file_types.get(ext, 0) + 1

        return file_types

    def _analyze_workflow(self, commits: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Analytical workflow patterns"""
        workflow = {
            "test_first": False,
            "commit_frequency": {},
            "pair_programming": False
        }

        # Check if there is a test-driven development model
        test_commits = [c for c in commits if any(
            "test" in f.lower() for f in c.get("files", [])
        )]

        if len(test_commits) > len(commits) * 0.3:
            workflow["test_first"] = True
            workflow["test_ratio"] = len(test_commits) / len(commits)

        return workflow

    def _calculate_confidence(self) -> float:
        """Calculate confidence"""
        total_commits = sum(self.preferences["tech_stack"].values())
        if total_commits < 10:
            return 0.3  # Not enough data，low confidence
        elif total_commits < 50:
            return 0.7  # medium confidence
        else:
            return 0.95  # high confidence


def main():
    """Test data collection"""
    collector = GitPreferenceCollector()
    preferences = collector.collect()

    print("\n📊 learning results：")
    print("Technology stack preferences：")
    for tech, count in sorted(preferences["tech_stack"].items(), key=lambda x: x[1], reverse=True):
        if count > 0:
            print(f"  - {tech}: {count}Second-rate")

    print("\nFile type preference：")
    for ext, count in sorted(preferences["file_types"].items(), key=lambda x: x[1], reverse=True)[:5]:
        print(f"  - {ext}: {count}Second-rate")

    print(f"\nConfidence: {preferences['metadata']['confidence']*100}%")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Automatic parameter adjustment system - Learning rate scheduling and hyperparameter search

Implement intelligent hyperparameter optimization：
- Learning rate scheduling（exponential decay、cosine annealing）
- Hyperparameter search（grid search、random search、Bayesian optimization）
- Early stopping mechanism（Early Stopping）
- Hyperparameter importance analysis

Phase 3.2 - expected250OK
"""

import numpy as np
from typing import Any, Callable, Dict, List, Optional, Tuple
from pathlib import Path
import json
from datetime import datetime
import logging
from copy import deepcopy

logger = logging.getLogger(__name__)


class LearningRateScheduler:
    """
    Learning rate scheduler

    Support multiple scheduling strategies：
    - constant: constant learning rate
    - exponential: exponential decay
    - cosine: cosine annealing
    - step: step decay
    """

    def __init__(self,
                 initial_lr: float = 0.001,
                 scheduler_type: str = "exponential",
                 decay_rate: float = 0.96,
                 decay_steps: int = 100,
                 min_lr: float = 1e-6):
        """
        Initialize the learning rate scheduler

        Args:
            initial_lr: Initial learning rate
            scheduler_type: scheduler type（constant/exponential/cosine/step）
            decay_rate: decay rate
            decay_steps: Decay steps
            min_lr: minimum learning rate
        """
        self.initial_lr = initial_lr
        self.scheduler_type = scheduler_type
        self.decay_rate = decay_rate
        self.decay_steps = decay_steps
        self.min_lr = min_lr

        self.current_step = 0
        self.history: List[Tuple[int, float]] = []

    def get_lr(self, step: int | None = None) -> float:
        """
        Get the current learning rate

        Args:
            step: number of steps（Nonethen use the internal count）

        Returns:
            Current learning rate
        """
        if step is None:
            step = self.current_step
            self.current_step += 1

        if self.scheduler_type == "constant":
            lr = self.initial_lr

        elif self.scheduler_type == "exponential":
            lr = self.initial_lr * (self.decay_rate ** (step / self.decay_steps))

        elif self.scheduler_type == "cosine":
            lr = self.min_lr + 0.5 * (self.initial_lr - self.min_lr) * (
                1 + np.cos(np.pi * step / self.decay_steps)
            )

        elif self.scheduler_type == "step":
            lr = self.initial_lr * (self.decay_rate ** (step // self.decay_steps))

        else:
            raise ValueError(f"Unknown scheduler type: {self.scheduler_type}")

        # Make sure not to go below the minimum learning rate
        lr = max(lr, self.min_lr)

        self.history.append((step, lr))
        return lr

    def reset(self):
        """reset scheduler"""
        self.current_step = 0
        self.history = []


class HyperparameterSearch:
    """
    Hyperparameter search

    Supports multiple search strategies：
    - grid: grid search（exhaustive）
    - random: random search
    - bayesian: Bayesian optimization（Simplified version）
    """

    def __init__(self,
                 search_space: Dict[str, Any],
                 search_type: str = "random",
                 n_trials: int = 50):
        """
        Initialize hyperparameter search

        Args:
            search_space: search space
                {
                    "learning_rate": (0.0001, 0.01),  # continuous range
                    "batch_size": [32, 64, 128],       # discrete choice
                    "hidden_dims": [[64, 64], [128, 128]]
                }
            search_type: Search type（grid/random/bayesian）
            n_trials: number of trials
        """
        self.search_space = search_space
        self.search_type = search_type
        self.n_trials = n_trials

        self.trials: List[Dict[str, Any]] = []
        self.best_trial: Optional[Dict[str, Any]] = None

    def suggest(self, trial_id: int) -> Dict[str, Any]:
        """
        Suggest next set of hyperparameters

        Args:
            trial_id: testID

        Returns:
            Hyperparameter configuration
        """
        if self.search_type == "grid":
            return self._grid_suggest(trial_id)
        elif self.search_type == "random":
            return self._random_suggest(trial_id)
        elif self.search_type == "bayesian":
            return self._bayesian_suggest(trial_id)
        else:
            raise ValueError(f"Unknown search type: {self.search_type}")

    def _grid_suggest(self, trial_id: int) -> Dict[str, Any]:
        """grid search"""
        # Generate all possible combinations（Simplified version）
        all_configs = self._generate_grid_configs()

        if trial_id >= len(all_configs):
            logger.warning(f"testID {trial_id} Grid search range exceeded，Use random configuration")
            return self._random_suggest(trial_id)

        return all_configs[trial_id]

    def _random_suggest(self, trial_id: int) -> Dict[str, Any]:
        """random search"""
        config = {}

        for param_name, param_space in self.search_space.items():
            if isinstance(param_space, tuple):
                # continuous range
                low, high = param_space
                value = np.random.uniform(low, high)
            elif isinstance(param_space, list):
                # discrete choice
                value = np.random.choice(param_space)
            else:
                raise ValueError(f"Unknown parameter space type: {type(param_space)}")

            config[param_name] = value

        return config

    def _bayesian_suggest(self, trial_id: int) -> Dict[str, Any]:
        """
        Bayesian optimization（Simplified version：Gaussian sampling based on historical results）

        Complete implementation needsskoptoroptunaLibrary，A simplified version is available here
        """
        if len(self.trials) < 5:
            # The first few experiments used random searches
            return self._random_suggest(trial_id)

        # Local search based on best trials
        if self.best_trial is None:
            return self._random_suggest(trial_id)

        best_config = self.best_trial["config"]
        config = {}

        for param_name, param_space in self.search_space.items():
            if isinstance(param_space, tuple):
                # continuous range：Sample around the optimal value
                low, high = param_space
                best_value = best_config[param_name]
                std = (high - low) * 0.1  # The standard deviation is the range10%

                value = np.random.normal(best_value, std)
                value = np.clip(value, low, high)
            elif isinstance(param_space, list):
                # discrete choice：have80%Probability selects the best value
                if np.random.random() < 0.8:
                    value = best_config[param_name]
                else:
                    value = np.random.choice(param_space)
            else:
                raise ValueError(f"Unknown parameter space type: {type(param_space)}")

            config[param_name] = value

        return config

    def _generate_grid_configs(self) -> List[Dict[str, Any]]:
        """Generate all configurations for grid search"""
        import itertools

        # Generate candidate values ​​for each parameter
        param_candidates = {}
        for param_name, param_space in self.search_space.items():
            if isinstance(param_space, tuple):
                # continuous range：generate5candidate values
                low, high = param_space
                param_candidates[param_name] = np.linspace(low, high, 5).tolist()
            elif isinstance(param_space, list):
                # discrete choice：use all values
                param_candidates[param_name] = param_space
            else:
                raise ValueError(f"Unknown parameter space type: {type(param_space)}")

        # Generate all combinations
        param_names = list(param_candidates.keys())
        param_values = [param_candidates[name] for name in param_names]

        all_configs = []
        for combination in itertools.product(*param_values):
            config = dict(zip(param_names, combination))
            all_configs.append(config)

        return all_configs

    def record_trial(
        self,
        config: Dict[str, Any],
        score: float,
        metadata: Dict[str, Any] | None = None,
    ) -> None:
        """
        Record test results

        Args:
            config: Hyperparameter configuration
            score: score
            metadata: additional metadata
        """
        trial = {
            "trial_id": len(self.trials),
            "config": deepcopy(config),
            "score": score,
            "timestamp": datetime.now().isoformat(),
            "metadata": metadata or {}
        }

        self.trials.append(trial)

        # Update best experiments
        if self.best_trial is None or score > self.best_trial["score"]:
            self.best_trial = trial
            logger.info(f"✅ new best test: {score:.4f}")

    def get_best_config(self) -> Optional[Dict[str, Any]]:
        """Get the best configuration"""
        if self.best_trial is None:
            return None
        return self.best_trial["config"]

    def analyze_importance(self) -> Dict[str, float]:
        """
        Analyze the importance of hyperparameters

        Use simple ANOVA（ANOVA）Thought

        Returns:
            Parameter importance score
        """
        if len(self.trials) < 10:
            logger.warning("⚠️ Not enough tests，Importance analysis cannot be performed")
            return {}

        importance: Dict[str, float] = {}

        for param_name in self.search_space.keys():
            # Group by parameter value
            groups: Dict[Any, List[float]] = {}
            for trial in self.trials:
                value = trial["config"][param_name]

                # Discretize continuous values
                if isinstance(value, float):
                    value = round(value, 4)

                if value not in groups:
                    groups[value] = []
                groups[value].append(trial["score"])

            # Calculate variance between groups
            if len(groups) < 2:
                importance[param_name] = 0.0
                continue

            group_means = [np.mean(scores) for scores in groups.values()]
            overall_mean = np.mean([trial["score"] for trial in self.trials])

            # sum of squares between groups
            ss_between = sum(
                len(groups[value]) * (mean - overall_mean) ** 2
                for value, mean in zip(groups.keys(), group_means)
            )

            # total sum of squares
            ss_total = sum(
                (trial["score"] - overall_mean) ** 2
                for trial in self.trials
            )

            # importance（proportion of explained variance）
            importance[param_name] = ss_between / ss_total if ss_total > 0 else 0.0

        # normalization
        total_importance = sum(importance.values())
        if total_importance > 0:
            importance = {k: v / total_importance for k, v in importance.items()}

        return importance


class EarlyStopping:
    """
    Early stopping mechanism

    Stop training when validation metrics no longer improve
    """

    def __init__(self,
                 patience: int = 10,
                 min_delta: float = 0.001,
                 mode: str = "max"):
        """
        Initialization early stop

        Args:
            patience: Number of steps tolerated（How many steps does it take to stop without improvement?）
            min_delta: minimum amount of improvement
            mode: model（maxBigger is better，minThe smaller the better）
        """
        self.patience = patience
        self.min_delta = min_delta
        self.mode = mode

        self.best_score: float | None = None
        self.counter = 0
        self.early_stop = False

        self.history: List[float] = []

    def check(self, score: float) -> bool:
        """
        Check if it should be stopped early

        Args:
            score: Current rating

        Returns:
            TrueIndicates it should stop
        """
        self.history.append(score)

        if self.best_score is None:
            self.best_score = score
            return False

        # Check if it improves
        if self.mode == "max":
            improved = score > self.best_score + self.min_delta
        else:
            improved = score < self.best_score - self.min_delta

        if improved:
            self.best_score = score
            self.counter = 0
        else:
            self.counter += 1

        # Check if tolerance is exceeded
        if self.counter >= self.patience:
            self.early_stop = True
            logger.info(f"⏹️ Early stop trigger（{self.counter}No improvement）")
            return True

        return False

    def reset(self):
        """Reset early stop"""
        self.best_score = None
        self.counter = 0
        self.early_stop = False
        self.history = []


class HyperparameterTuner:
    """
    Hyperparameter tuner（Integrate all functions）

    Complete automatic parameter adjustment process：
    1. Learning rate scheduling
    2. Hyperparameter search
    3. Early stopping mechanism
    4. Result analysis
    """

    def __init__(self,
                 search_space: Dict[str, Any],
                 search_type: str = "random",
                 n_trials: int = 50,
                 patience: int = 10):
        """
        Initialize the tuner

        Args:
            search_space: search space
            search_type: Search type
            n_trials: number of trials
            patience: Early stopping tolerance steps
        """
        self.search = HyperparameterSearch(search_space, search_type, n_trials)
        self.early_stopping = EarlyStopping(patience=patience)
        self.results: List[Dict[str, Any]] = []

    def optimize(
        self,
        objective_fn: Callable[[Dict[str, Any]], float],
        save_dir: str | Path | None = None,
    ) -> Dict[str, Any]:
        """
        Perform hyperparameter optimization

        Args:
            objective_fn: objective function（Receive hyperparameter configuration，Return to rating）
            save_dir: Result saving directory

        Returns:
            Optimize results
        """
        logger.info(f"🔍 Start hyperparameter search（{self.search.n_trials} trials）...")

        save_path = Path(save_dir) if save_dir else Path("./hyperparameter_tuning_results")
        save_path.mkdir(parents=True, exist_ok=True)

        for trial_id in range(self.search.n_trials):
            # Recommended configuration
            config = self.search.suggest(trial_id)

            logger.info(f"🎲 test {trial_id + 1}/{self.search.n_trials}")
            logger.info(f"   Configuration: {json.dumps(config, indent=2)}")

            try:
                # Evaluate configuration
                score = objective_fn(config)

                # Record results
                self.search.record_trial(config, score)
                self.early_stopping.check(score)

                self.results.append({
                    "trial_id": trial_id,
                    "config": config,
                    "score": score
                })

                logger.info(f"   score: {score:.4f}")

                # Save intermediate results
                self._save_intermediate_results(save_path, trial_id)

                # Inspection stopped early
                if self.early_stopping.early_stop:
                    logger.info("⏹️ Early stop trigger，End search")
                    break

            except Exception as e:
                logger.error(f"❌ test {trial_id} fail: {e}")
                continue

        # Analyze results
        return self._analyze_results(save_path)

    def _save_intermediate_results(self, save_path: Path, trial_id: int):
        """Save intermediate results"""
        results_file = save_path / f"intermediate_results_trial_{trial_id}.json"

        with open(results_file, 'w') as f:
            json.dump({
                "trial_id": trial_id,
                "trials": self.search.trials,
                "best_trial": self.search.best_trial
            }, f, indent=2)

    def _analyze_results(self, save_path: Path) -> Dict[str, Any]:
        """Analyze optimization results"""
        # Get the best configuration
        best_config = self.search.get_best_config()
        best_score = self.search.best_trial["score"] if self.search.best_trial else None

        # Analyze parameter importance
        importance = self.search.analyze_importance()

        # Save final result
        final_results = {
            "best_config": best_config,
            "best_score": best_score,
            "parameter_importance": importance,
            "all_trials": self.search.trials,
            "convergence_curve": [t["score"] for t in self.search.trials]
        }

        results_file = save_path / "final_results.json"
        with open(results_file, 'w') as f:
            json.dump(final_results, f, indent=2)

        logger.info("✅ Optimization completed！")
        logger.info(f"   best rating: {best_score:.4f}")
        logger.info(f"   Best configuration: {json.dumps(best_config, indent=2)}")
        logger.info(f"   Parameter importance: {json.dumps(importance, indent=2)}")
        logger.info(f"   Results saved: {results_file}")

        return final_results


def main():
    """Testing the hyperparameter tuner"""
    # Define the search space
    search_space = {
        "learning_rate": (0.0001, 0.01),
        "batch_size": [32, 64, 128],
        "hidden_dims": [[64, 64], [128, 128], [256, 256]]
    }

    # Create a tuner
    tuner = HyperparameterTuner(
        search_space=search_space,
        search_type="random",
        n_trials=10,
        patience=5
    )

    # Define objective function（simulation）
    def objective_fn(config):
        # Simulation training process
        lr = config["learning_rate"]
        batch_size = config["batch_size"]

        # mock scoring（The closer to the optimal value, the higher）
        optimal_lr = 0.001
        optimal_batch = 64

        lr_score = 1.0 - abs(lr - optimal_lr) / optimal_lr
        batch_score = 1.0 - abs(batch_size - optimal_batch) / optimal_batch

        return (lr_score + batch_score) / 2 + np.random.normal(0, 0.05)

    # Perform optimization
    results = tuner.optimize(objective_fn, save_dir="./tuning_test")

    print("\n📊 Optimize results:")
    print(f"   best rating: {results['best_score']:.4f}")
    print(f"   Parameter importance: {results['parameter_importance']}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Canonical policy models for rule, playbook, and policy event assets."""

from __future__ import annotations

import hashlib
import json
from dataclasses import asdict, dataclass, field
from typing import Any

RULE_STATUSES = {"hint", "candidate", "confirmed", "suspended", "archived"}
RULE_SCOPES = {"global", "domain", "project"}
POLICY_DECISIONS = {"", "auto_execute", "require_confirmation"}


@dataclass
class Rule:
    """A reusable confirmation rule kept in local policy memory."""

    id: str
    type: str = "Rule"
    summary: str = ""
    category: str = "innovate"
    trigger: list[str] = field(default_factory=list)
    strategy: str = ""
    preconditions: list[str] = field(default_factory=list)
    postconditions: list[str] = field(default_factory=list)
    validation: list[str] = field(default_factory=list)
    confidence: float = 0.0
    success_streak: int = 0
    status: str = "hint"
    scope: str = "global"
    scope_key: str = ""
    evidence_count: int = 0
    source_type: str = "legacy"
    last_seen_at: str = ""
    last_applied_at: str = ""
    policy_decision: str = ""
    risk_level: str = ""
    failure_streak: int = 0
    override_streak: int = 0
    rollback_count: int = 0
    accepted_auto_execute_count: int = 0
    accepted_confirmation_count: int = 0
    conflict_count: int = 0
    suspension_reason: str = ""
    last_lifecycle_event_at: str = ""
    asset_id: str = ""

    def calculate_asset_id(self) -> str:
        """Compute a deterministic asset id from record content."""
        content = asdict(self)
        content.pop("asset_id", None)
        serialized = json.dumps(content, sort_keys=True, ensure_ascii=False)
        digest = hashlib.sha256(serialized.encode("utf-8")).hexdigest()
        self.asset_id = f"sha256:{digest}"
        return self.asset_id

    def to_dict(self) -> dict[str, Any]:
        """Serialize to a dictionary."""
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Rule":
        """Deserialize and normalize legacy payloads."""
        payload = dict(data)
        payload["type"] = "Rule"
        payload.setdefault("status", "hint")
        payload.setdefault("scope", "global")
        payload.setdefault("scope_key", "")
        payload.setdefault("evidence_count", 0)
        payload.setdefault("source_type", "legacy")
        payload.setdefault("last_seen_at", "")
        payload.setdefault("last_applied_at", "")
        payload.setdefault("policy_decision", "")
        payload.setdefault("risk_level", "")
        payload.setdefault("failure_streak", 0)
        payload.setdefault("override_streak", 0)
        payload.setdefault("rollback_count", 0)
        payload.setdefault("accepted_auto_execute_count", 0)
        payload.setdefault("accepted_confirmation_count", 0)
        payload.setdefault("conflict_count", 0)
        payload.setdefault("suspension_reason", "")
        payload.setdefault("last_lifecycle_event_at", "")

        if payload["status"] not in RULE_STATUSES:
            payload["status"] = "hint"
        if payload["scope"] not in RULE_SCOPES:
            payload["scope"] = "global"
        if payload["policy_decision"] not in POLICY_DECISIONS:
            payload["policy_decision"] = ""

        return cls(**payload)

    def increment_confidence(self, reward: float) -> None:
        """Adjust confidence and success streak from a reward signal."""
        if reward > 0.7:
            self.success_streak += 1
            self.confidence = min(1.0, self.confidence + 0.05)
        elif reward < 0.3:
            self.success_streak = 0
            self.confidence = max(0.0, self.confidence - 0.1)

    def __str__(self) -> str:
        return f"[{self.id}] {self.summary} (confidence: {self.confidence:.2f})"


@dataclass
class Playbook:
    """A reusable playbook composed of multiple confirmation rules."""

    id: str
    type: str = "Playbook"
    summary: str = ""
    rules_used: list[str] = field(default_factory=list)
    trigger: list[str] = field(default_factory=list)
    category: str = "innovate"
    confidence: float = 0.0
    asset_id: str = ""

    def calculate_asset_id(self) -> str:
        """Compute a deterministic asset id from record content."""
        content = asdict(self)
        content.pop("asset_id", None)
        serialized = json.dumps(content, sort_keys=True, ensure_ascii=False)
        digest = hashlib.sha256(serialized.encode("utf-8")).hexdigest()
        self.asset_id = f"sha256:{digest}"
        return self.asset_id

    def to_dict(self) -> dict[str, Any]:
        """Serialize to a dictionary."""
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Playbook":
        """Deserialize and normalize legacy payloads."""
        payload = dict(data)
        payload["type"] = "Playbook"
        payload["rules_used"] = list(payload.get("rules_used") or payload.get("genes_used") or [])
        payload.pop("genes_used", None)
        return cls(**payload)

    def __str__(self) -> str:
        return (
            f"[{self.id}] {self.summary} "
            f"({len(self.rules_used)} rules, confidence: {self.confidence:.2f})"
        )


@dataclass
class PolicyEvent:
    """Append-only policy event for evaluation and decision history."""

    timestamp: str
    event_type: str
    asset_id: str
    type: str = "PolicyEvent"
    trigger_signals: list[str] = field(default_factory=list)
    rl_reward: float = 0.0
    changes: str = ""
    source_node_id: str = ""
    payload: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        """Serialize to a dictionary."""
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "PolicyEvent":
        """Deserialize and normalize legacy payloads."""
        payload = dict(data)
        payload["type"] = "PolicyEvent"
        return cls(**payload)

    def to_jsonl(self) -> str:
        """Serialize as one JSONL line."""
        return json.dumps(asdict(self), ensure_ascii=False)

    @classmethod
    def from_jsonl(cls, jsonl_str: str) -> "PolicyEvent":
        """Deserialize from one JSONL line."""
        return cls.from_dict(json.loads(jsonl_str))

    def __str__(self) -> str:
        if self.payload.get("decision_id"):
            decision = self.payload.get("final_decision", "unknown")
            return f"[{self.timestamp}] {self.event_type}: {decision} ({self.payload['decision_id']})"
        return f"[{self.timestamp}] {self.event_type}: {self.changes} (reward: {self.rl_reward:.2f})"

__all__ = [
    "RULE_STATUSES",
    "RULE_SCOPES",
    "POLICY_DECISIONS",
    "Rule",
    "Playbook",
    "PolicyEvent",
]



#!/usr/bin/env python3
"""Unit tests for canonical policy models and storage manager."""

from __future__ import annotations

import tempfile
from pathlib import Path
from shutil import rmtree

import pytest

from lib.policy_models import Playbook, PolicyEvent, Rule
from lib.policy_store import PolicyStore


class TestRule:
    def test_rule_creation(self) -> None:
        rule = Rule(
            id="test_rule",
            summary="Test rule",
            category="optimize",
            strategy="Test strategy",
            confidence=0.8,
        )

        assert rule.id == "test_rule"
        assert rule.type == "Rule"
        assert rule.summary == "Test rule"
        assert rule.category == "optimize"
        assert rule.confidence == 0.8

    def test_rule_calculate_asset_id(self) -> None:
        rule = Rule(
            id="test_rule",
            summary="Test rule",
            category="optimize",
            strategy="Test strategy",
        )

        asset_id = rule.calculate_asset_id()

        assert asset_id.startswith("sha256:")
        assert len(asset_id) == 71

    def test_rule_to_dict(self) -> None:
        rule = Rule(id="test_rule", summary="Test rule", category="optimize", confidence=0.8)
        data = rule.to_dict()

        assert isinstance(data, dict)
        assert data["id"] == "test_rule"
        assert data["summary"] == "Test rule"

    def test_rule_from_dict(self) -> None:
        data = {
            "id": "test_rule",
            "type": "Rule",
            "summary": "Test rule",
            "category": "optimize",
            "trigger": [],
            "strategy": "",
            "preconditions": [],
            "postconditions": [],
            "validation": [],
            "confidence": 0.8,
            "success_streak": 0,
            "asset_id": "",
        }

        rule = Rule.from_dict(data)

        assert rule.id == "test_rule"
        assert rule.type == "Rule"
        assert rule.summary == "Test rule"
        assert rule.confidence == 0.8

    def test_rule_from_dict_applies_policy_defaults(self) -> None:
        rule = Rule.from_dict(
            {
                "id": "legacy_rule",
                "summary": "Legacy rule",
                "trigger": ["T2"],
                "confidence": 0.9,
                "success_streak": 6,
            }
        )

        assert rule.status == "hint"
        assert rule.scope == "global"
        assert rule.scope_key == ""
        assert rule.evidence_count == 0
        assert rule.source_type == "legacy"
        assert rule.policy_decision == ""

    def test_rule_increment_confidence(self) -> None:
        rule = Rule(id="test_rule", summary="Test rule", confidence=0.5, success_streak=0)

        rule.increment_confidence(0.8)
        assert rule.success_streak == 1
        assert rule.confidence > 0.5

        previous_confidence = rule.confidence
        rule.increment_confidence(0.2)
        assert rule.success_streak == 0
        assert rule.confidence < previous_confidence


class TestPlaybook:
    def test_playbook_creation(self) -> None:
        playbook = Playbook(
            id="test_playbook",
            summary="Test playbook",
            rules_used=["rule1", "rule2"],
            category="innovate",
        )

        assert playbook.id == "test_playbook"
        assert playbook.type == "Playbook"
        assert len(playbook.rules_used) == 2

    def test_playbook_calculate_asset_id(self) -> None:
        playbook = Playbook(id="test_playbook", summary="Test playbook", rules_used=[])

        asset_id = playbook.calculate_asset_id()

        assert asset_id.startswith("sha256:")
        assert len(asset_id) == 71


class TestPolicyEvent:
    def test_policy_event_creation(self) -> None:
        event = PolicyEvent(
            timestamp="2026-03-01T12:00:00Z",
            event_type="rule_created",
            asset_id="sha256:abc123",
            rl_reward=0.8,
        )

        assert event.type == "PolicyEvent"
        assert event.event_type == "rule_created"
        assert event.rl_reward == 0.8

    def test_policy_event_to_jsonl(self) -> None:
        event = PolicyEvent(
            timestamp="2026-03-01T12:00:00Z",
            event_type="rule_created",
            asset_id="sha256:abc123",
            changes="Created new rule",
        )

        jsonl = event.to_jsonl()

        assert isinstance(jsonl, str)
        assert "rule_created" in jsonl
        assert "Created new rule" in jsonl

    def test_policy_event_from_jsonl(self) -> None:
        jsonl = (
            '{"timestamp":"2026-03-01T12:00:00Z","event_type":"rule_created",'
            '"asset_id":"sha256:abc123","trigger_signals":[],"rl_reward":0.8,'
            '"changes":"Created new rule","source_node_id":"test"}'
        )

        event = PolicyEvent.from_jsonl(jsonl)

        assert event.type == "PolicyEvent"
        assert event.event_type == "rule_created"
        assert event.changes == "Created new rule"
        assert event.rl_reward == 0.8

    def test_policy_event_payload_roundtrip(self) -> None:
        event = PolicyEvent(
            timestamp="2026-03-01T12:00:00Z",
            event_type="decision_evaluated",
            asset_id="sha256:decision",
            payload={
                "decision_id": "dec_123",
                "final_decision": "auto_execute",
            },
        )

        restored = PolicyEvent.from_jsonl(event.to_jsonl())

        assert restored.payload["decision_id"] == "dec_123"
        assert restored.payload["final_decision"] == "auto_execute"


class TestPolicyStore:
    @pytest.fixture
    def temp_dir(self) -> Path:
        temp_dir = Path(tempfile.mkdtemp())
        yield temp_dir
        if temp_dir.exists():
            rmtree(temp_dir)

    def test_policy_store_init(self, temp_dir: Path) -> None:
        store = PolicyStore(temp_dir)

        assert store.base_dir == temp_dir
        assert store.rules_file == temp_dir / "rules.json"
        assert store.playbooks_file == temp_dir / "playbooks.json"
        assert store.policy_events_file == temp_dir / "policy_events.jsonl"

    def test_save_and_load_rules(self, temp_dir: Path) -> None:
        store = PolicyStore(temp_dir)
        rule = Rule(id="test_rule", summary="Test rule", category="optimize", confidence=0.8)
        rule.calculate_asset_id()

        store.save_rule(rule)
        loaded_rule = store.get_rule("test_rule")

        assert loaded_rule is not None
        assert loaded_rule.id == "test_rule"
        assert loaded_rule.summary == "Test rule"
        assert loaded_rule.confidence == 0.8

    def test_save_and_load_playbooks(self, temp_dir: Path) -> None:
        store = PolicyStore(temp_dir)
        playbook = Playbook(id="test_playbook", summary="Test playbook", rules_used=["rule1"])
        playbook.calculate_asset_id()

        store.save_playbook(playbook)
        loaded_playbook = store.get_playbook("test_playbook")

        assert loaded_playbook is not None
        assert loaded_playbook.id == "test_playbook"
        assert loaded_playbook.summary == "Test playbook"

    def test_append_and_get_events(self, temp_dir: Path) -> None:
        store = PolicyStore(temp_dir)

        event1 = PolicyEvent(
            timestamp="2026-03-01T12:00:00Z",
            event_type="rule_created",
            asset_id="sha256:abc123",
            changes="Created rule 1",
        )

        event2 = PolicyEvent(
            timestamp="2026-03-01T12:05:00Z",
            event_type="rule_created",
            asset_id="sha256:def456",
            changes="Created rule 2",
        )

        store.append_event(event1)
        store.append_event(event2)

        events = store.get_events(limit=10)

        assert len(events) == 2
        assert events[0].changes == "Created rule 2"
        assert events[1].changes == "Created rule 1"

    def test_delete_rule(self, temp_dir: Path) -> None:
        store = PolicyStore(temp_dir)

        rule = Rule(id="test_rule", summary="Test rule")
        store.save_rule(rule)

        result = store.delete_rule("test_rule")

        assert result is True
        assert store.get_rule("test_rule") is None



#!/usr/bin/env python3
"""
Monitoring panel - TensorBoardIntegration and training visualization

Realize complete training monitoring function：
- TensorBoardlogging
- Reward Curve Visualization
- loss curve tracking
- Hyperparameter comparison
- Model performance analysis

Phase 3.3 - expected180OK
"""

import numpy as np
from typing import Any, Dict, List, Tuple
from pathlib import Path
import json
from datetime import datetime
import logging

logger = logging.getLogger(__name__)

# try to importTensorBoard
try:
    from torch.utils import tensorboard as _tensorboard
    _SummaryWriterFactory: Any = _tensorboard.SummaryWriter
    TENSORBOARD_AVAILABLE = True
except ImportError:
    TENSORBOARD_AVAILABLE = False
    _SummaryWriterFactory = None


class TrainingMonitor:
    """
    training monitor

    Function：
    - TensorBoardlogging
    - Metric tracking（award、loss、performance）
    - Checkpoint comparison
    - Training report generation
    """

    def __init__(self,
                 log_dir: str = "./logs/tensorboard",
                 experiment_name: str | None = None):
        """
        Initialize monitor

        Args:
            log_dir: Log directory
            experiment_name: Experiment name
        """
        self.log_dir = Path(log_dir).expanduser()
        self.log_dir.mkdir(parents=True, exist_ok=True)

        # Create experiment directory
        if experiment_name:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            self.experiment_dir = self.log_dir / f"{experiment_name}_{timestamp}"
        else:
            self.experiment_dir = self.log_dir / datetime.now().strftime("%Y%m%d_%H%M%S")

        self.experiment_dir.mkdir(parents=True, exist_ok=True)

        # initializationTensorBoard writer
        self.tensorboard_enabled = TENSORBOARD_AVAILABLE
        self.writer: Any | None = None
        if self.tensorboard_enabled:
            self.writer = _SummaryWriterFactory(log_dir=str(self.experiment_dir))
            logger.info(f"✅ TensorBoardEnabled: {self.experiment_dir}")
        else:
            logger.warning("⚠️ TensorBoardNot available，Use degraded mode（JSONlog）")

        # Indicator record
        self.metrics_history: Dict[str, List[Tuple[int, float]]] = {}
        self._closed = False

    def log_scalar(self, tag: str, value: float, step: int):
        """
        Logging scalar metrics

        Args:
            tag: Indicator label（like"train/reward", "train/loss"）
            value: indicator value
            step: number of steps
        """
        if self.tensorboard_enabled and self.writer:
            self.writer.add_scalar(tag, value, step)

        # Record to memory simultaneously（for downgrade mode）
        if tag not in self.metrics_history:
            self.metrics_history[tag] = []
        self.metrics_history[tag].append((step, value))

    def log_scalars(self, main_tag: str, tag_scalar_dict: Dict[str, float], step: int):
        """
        Logging scalar metrics in batches

        Args:
            main_tag: main label
            tag_scalar_dict: Label-value dictionary
            step: number of steps
        """
        if self.tensorboard_enabled and self.writer:
            self.writer.add_scalars(main_tag, tag_scalar_dict, step)

        # Record to memory
        for sub_tag, value in tag_scalar_dict.items():
            full_tag = f"{main_tag}/{sub_tag}"
            if full_tag not in self.metrics_history:
                self.metrics_history[full_tag] = []
            self.metrics_history[full_tag].append((step, value))

    def log_histogram(self, tag: str, values: np.ndarray, step: int):
        """
        Record histogram

        Args:
            tag: Label
            values: value array
            step: number of steps
        """
        if self.tensorboard_enabled and self.writer:
            self.writer.add_histogram(tag, values, step)

    def log_hyperparams(self, hparam_dict: Dict[str, Any], metric_dict: Dict[str, float]):
        """
        Record hyperparameters and final metrics

        Args:
            hparam_dict: Hyperparameter dictionary
            metric_dict: indicator dictionary
        """
        if self.tensorboard_enabled and self.writer:
            self.writer.add_hparams(hparam_dict, metric_dict)

    def log_training_step(self,
                         episode: int,
                         reward: float,
                         actor_loss: float,
                         critic_loss: float,
                         extra_metrics: Dict[str, float] | None = None):
        """
        Record training steps（Convenience method）

        Args:
            episode: Episodenumber
            reward: award
            actor_loss: Actorloss
            critic_loss: Criticloss
            extra_metrics: additional indicators
        """
        # Record key indicators
        self.log_scalar("train/reward", reward, episode)
        self.log_scalar("train/actor_loss", actor_loss, episode)
        self.log_scalar("train/critic_loss", critic_loss, episode)

        # Record additional metrics
        if extra_metrics:
            for key, value in extra_metrics.items():
                self.log_scalar(f"train/{key}", value, episode)

    def log_evaluation(self,
                      episode: int,
                      eval_reward: float,
                      eval_success_rate: float,
                      eval_metrics: Dict[str, float] | None = None):
        """
        Record evaluation metrics

        Args:
            episode: Episodenumber
            eval_reward: Evaluation rewards
            eval_success_rate: success rate
            eval_metrics: Additional evaluation metrics
        """
        self.log_scalar("eval/reward", eval_reward, episode)
        self.log_scalar("eval/success_rate", eval_success_rate, episode)

        if eval_metrics:
            for key, value in eval_metrics.items():
                self.log_scalar(f"eval/{key}", value, episode)

    def log_model_weights(self, tag: str, weights: np.ndarray, step: int):
        """
        Record model weight distribution

        Args:
            tag: Label（like"policy_net/layer1"）
            weights: weight array
            step: number of steps
        """
        self.log_histogram(tag, weights, step)

        # Record statistics
        self.log_scalars(
            f"{tag}/stats",
            {
                "mean": float(np.mean(weights)),
                "std": float(np.std(weights)),
                "min": float(np.min(weights)),
                "max": float(np.max(weights))
            },
            step
        )

    def save_metrics_to_json(self, filename: str | None = None) -> str:
        """
        Save the indicator toJSONdocument（downgrade mode）

        Args:
            filename: file name

        Returns:
            file path
        """
        if not filename:
            filename = f"metrics_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"

        filepath = self.experiment_dir / filename

        with open(filepath, 'w') as f:
            json.dump(self.metrics_history, f, indent=2)

        logger.info(f"✅ Indicator saved: {filepath}")
        return str(filepath)

    def get_metrics_summary(self) -> Dict[str, Any]:
        """
        Get indicator summary

        Returns:
            Indicator summary statistics
        """
        summary = {}

        for tag, history in self.metrics_history.items():
            if not history:
                continue

            steps, values = zip(*history)

            summary[tag] = {
                "count": len(values),
                "mean": float(np.mean(values)),
                "std": float(np.std(values)),
                "min": float(np.min(values)),
                "max": float(np.max(values)),
                "latest": float(values[-1]),
                "first": float(values[0]),
                "improvement": float(values[-1] - values[0]) if len(values) > 1 else 0.0
            }

        return summary

    def plot_metrics(self, tags: List[str] | None = None, save_path: str | Path | None = None):
        """
        Draw indicator curve（downgrade mode）

        Args:
            tags: Indicator labels to plot（None=all）
            save_path: save path
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            logger.warning("⚠️ matplotlibNot available，Unable to draw")
            return

        if tags is None:
            tags = list(self.metrics_history.keys())

        # Create subgraph
        n_plots = len(tags)
        n_cols = min(3, n_plots)
        n_rows = (n_plots + n_cols - 1) // n_cols

        fig, axes = plt.subplots(n_rows, n_cols, figsize=(15, 5 * n_rows))
        if n_plots == 1:
            axes = [axes]
        else:
            axes = axes.flatten()

        # Plot each indicator
        for i, tag in enumerate(tags):
            if tag not in self.metrics_history:
                continue

            steps, values = zip(*self.metrics_history[tag])

            axes[i].plot(steps, values, 'b-', linewidth=2)
            axes[i].set_title(tag)
            axes[i].set_xlabel('Step')
            axes[i].set_ylabel('Value')
            axes[i].grid(True, alpha=0.3)

        # Hide redundant subplots
        for i in range(n_plots, len(axes)):
            axes[i].set_visible(False)

        plt.tight_layout()

        plot_path = Path(save_path).expanduser() if save_path else self.experiment_dir / "metrics_plot.png"
        plt.savefig(plot_path, dpi=150, bbox_inches='tight')
        logger.info(f"✅ Chart saved: {plot_path}")

        plt.close()

    def compare_experiments(self,
                          experiment_dirs: List[str],
                          metric_tags: List[str]) -> Dict[str, Any]:
        """
        Compare multiple experiments

        Args:
            experiment_dirs: Experiment directory list
            metric_tags: Metric labels to compare

        Returns:
            Compare results
        """
        comparison: Dict[str, Any] = {
            "experiments": [],
            "metrics": {}
        }

        for exp_dir in experiment_dirs:
            exp_path = Path(exp_dir)
            if not exp_path.exists():
                logger.warning(f"⚠️ Experiment directory does not exist: {exp_dir}")
                continue

            # Read indicator file
            metrics_file = exp_path / "metrics_summary.json"
            if metrics_file.exists():
                with open(metrics_file, 'r') as f:
                    metrics = json.load(f)
                    comparison["experiments"].append({
                        "name": exp_path.name,
                        "metrics": metrics
                    })

        # Calculate the best experiment for each metric
        for tag in metric_tags:
            best_exp = None
            best_value = None

            for exp in comparison["experiments"]:
                if tag in exp["metrics"]:
                    value = exp["metrics"][tag].get("latest", 0)

                    if best_value is None or value > best_value:
                        best_value = value
                        best_exp = exp["name"]

            comparison["metrics"][tag] = {
                "best_experiment": best_exp,
                "best_value": best_value
            }

        return comparison

    def close(self):
        """Close monitor"""
        if self._closed:
            return
        self._safe_close_writer()
        self._closed = True

        # Save final indicator
        summary = self.get_metrics_summary()
        summary_file = self.experiment_dir / "metrics_summary.json"
        with open(summary_file, 'w') as f:
            json.dump(summary, f, indent=2)

        logger.info("✅ Monitor is off，Indicator saved")

    def _safe_close_writer(self):
        """Safe shutdownwriter，Avoid thread exceptions"""
        if self.tensorboard_enabled and self.writer:
            try:
                self.writer.flush()
                self.writer.close()
                logger.info("✅ TensorBoard writerClosed")
            except Exception as e:
                logger.warning(f"⚠️ closureTensorBoard writerfail: {e}")

    def __del__(self):
        """Try to close when destructingwriter，Avoid background thread exceptions"""
        try:
            self.close()
        except Exception:
            # Avoid exceptions thrown during destruction
            pass


class MetricsAnalyzer:
    """
    Indicator Analyzer

    Analyze training metrics and provide insights
    """

    def __init__(self, metrics_history: Dict[str, List[Tuple[int, float]]]):
        """
        Initialize analyzer

        Args:
            metrics_history: Indicator history
        """
        self.metrics = metrics_history

    def detect_convergence(self,
                          metric_tag: str,
                          window: int = 10,
                          threshold: float = 0.01) -> Tuple[bool, int]:
        """
        Check whether the indicator converges

        Args:
            metric_tag: Indicator label
            window: sliding window size
            threshold: convergence threshold（variance）

        Returns:
            (Convergence or not, Convergence steps)
        """
        if metric_tag not in self.metrics:
            return False, -1

        steps, values = zip(*self.metrics[metric_tag])

        # Sliding window detection
        for i in range(window, len(values)):
            window_values = values[i-window:i]
            variance = np.var(window_values)

            if variance < threshold:
                return True, steps[i]

        return False, -1

    def detect_plateau(self,
                      metric_tag: str,
                      window: int = 20,
                      threshold: float = 0.001) -> Tuple[bool, int]:
        """
        Detect whether the indicator has entered the plateau period

        Args:
            metric_tag: Indicator label
            window: sliding window size
            threshold: plateau threshold（average rate of change）

        Returns:
            (Is it a plateau?, Start steps)
        """
        if metric_tag not in self.metrics:
            return False, -1

        steps, values = zip(*self.metrics[metric_tag])

        # Calculate rate of change
        for i in range(window, len(values)):
            recent_values = values[i-window:i]
            avg_change = np.mean(np.diff(recent_values))

            if abs(avg_change) < threshold:
                return True, steps[i-window]

        return False, -1

    def analyze_learning_curve(self, metric_tag: str) -> Dict[str, Any]:
        """
        Analyze the learning curve

        Args:
            metric_tag: Indicator label

        Returns:
            Analyze results
        """
        if metric_tag not in self.metrics:
            return {}

        steps, values = zip(*self.metrics[metric_tag])

        # Fit the learning curve（exponential saturation model）
        def model(x, a, b, c):
            return a * (1 - np.exp(-b * x)) + c

        try:
            from scipy.optimize import curve_fit

            # Normalization steps
            x_normalized = np.array(steps) / max(steps)
            y = np.array(values)

            # Initial parameter guessing
            p0 = [max(y) - min(y), 1.0, min(y)]

            # fitting
            popt, _ = curve_fit(model, x_normalized, y, p0=p0, maxfev=5000)

            # Calculate fit quality
            y_pred = model(x_normalized, *popt)
            r_squared = 1 - np.sum((y - y_pred)**2) / np.sum((y - np.mean(y))**2)

            # Predict asymptotic value
            asymptote = popt[0] + popt[2]

            return {
                "convergence_value": float(asymptote),
                "current_value": float(values[-1]),
                "progress_to_convergence": float(values[-1] / asymptote) if asymptote != 0 else 0,
                "r_squared": float(r_squared),
                "fit_params": {
                    "amplitude": float(popt[0]),
                    "rate": float(popt[1]),
                    "offset": float(popt[2])
                }
            }

        except (ImportError, RuntimeError):
            # scipyNot available or fitting failed，Return simple statistics
            return {
                "current_value": float(values[-1]),
                "mean": float(np.mean(values)),
                "std": float(np.std(values)),
                "trend": "increasing" if values[-1] > values[0] else "decreasing"
            }


def main():
    """test monitor"""
    monitor = TrainingMonitor(experiment_name="test_monitor")

    print("✅ Monitor created")
    print(f"   TensorBoard: {'✅ enable' if monitor.tensorboard_enabled else '❌ Disable'}")

    # Simulation training process
    for episode in range(100):
        reward = 0.5 + 0.3 * (1 - np.exp(-episode / 50)) + np.random.normal(0, 0.02)
        actor_loss = 0.5 * np.exp(-episode / 30) + np.random.normal(0, 0.01)
        critic_loss = 0.3 * np.exp(-episode / 40) + np.random.normal(0, 0.01)

        monitor.log_training_step(
            episode=episode,
            reward=reward,
            actor_loss=actor_loss,
            critic_loss=critic_loss
        )

    # Save indicator
    monitor.save_metrics_to_json()

    # draw curve
    monitor.plot_metrics()

    # Get summary
    summary = monitor.get_metrics_summary()
    print("\n📊 Training summary:")
    print(f"   final reward: {summary['train/reward']['latest']:.4f}")
    print(f"   Reward improvements: {summary['train/reward']['improvement']:.4f}")
    print(f"   finalActorloss: {summary['train/actor_loss']['latest']:.4f}")

    # Analyze the learning curve
    analyzer = MetricsAnalyzer(monitor.metrics_history)
    curve_analysis = analyzer.analyze_learning_curve("train/reward")
    print("\n📈 Learning curve analysis:")
    print(f"   Convergence value prediction: {curve_analysis.get('convergence_value', 0):.4f}")
    print(f"   Current progress: {curve_analysis.get('progress_to_convergence', 0)*100:.1f}%")

    monitor.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Converter from Markdown memory files to policy memory assets."""

from __future__ import annotations

import re
import time
from pathlib import Path

from .policy_models import Playbook, PolicyEvent, Rule
from .policy_store import PolicyStore

_BASIC_INFO_SECTIONS = ("Basic Information", "\u57fa\u672c\u4fe1\u606f")
_WORK_PREF_SECTIONS = ("Working Preferences", "\u5de5\u4f5c\u504f\u597d")
_NOTES_SECTIONS = ("Notes", "\u5907\u6ce8")

_TOOL_DISPATCH_SECTIONS = ("Tool Dispatch", "\u5de5\u5177\u8c03\u5ea6")
_OPERATION_RULES_SECTIONS = ("Operation Rules", "\u64cd\u4f5c\u89c4\u5219")
_ESCALATION_SECTIONS = ("Escalation", "\u5347\u7ea7\u7b56\u7565")

_CORE_PRINCIPLES_SECTIONS = ("Core Principles", "\u6838\u5fc3\u539f\u5219")
_PROHIBITED_SECTIONS = ("Prohibited Actions", "\u7981\u6b62\u884c\u4e3a")
_REWARD_SECTIONS = ("Reward Signals", "\u5956\u52b1\u4fe1\u53f7")


class MarkdownToPolicyConverter:
    """Convert USER/SOUL/AGENTS markdown files into policy assets."""

    def convert_user_md_to_rules(self, user_md_path: Path) -> dict[str, Rule]:
        if not user_md_path.exists():
            return {}

        content = user_md_path.read_text(encoding="utf-8")
        rules: dict[str, Rule] = {}
        timestamp = int(time.time())

        basic_info = self._extract_section_by_aliases(content, _BASIC_INFO_SECTIONS)
        if basic_info:
            rules["rule_basic_info"] = Rule(
                id=f"rule_basic_info_{timestamp}",
                summary="Basic information",
                category="optimize",
                strategy=basic_info,
                trigger=["task_start", "user_query"],
                confidence=0.8,
            )

        work_prefs = self._extract_section_by_aliases(content, _WORK_PREF_SECTIONS)
        if work_prefs:
            rules["rule_work_habits"] = Rule(
                id=f"rule_work_habits_{timestamp}",
                summary="Working preferences",
                category="optimize",
                strategy=work_prefs,
                trigger=["task_execution", "workflow_decision"],
                confidence=0.8,
            )

        notes = self._extract_section_by_aliases(content, _NOTES_SECTIONS)
        if notes:
            rules["rule_project_constraints"] = Rule(
                id=f"rule_project_constraints_{timestamp}",
                summary="Project constraints",
                category="harden",
                strategy=notes,
                trigger=["constraint_check", "validation"],
                confidence=0.7,
            )

        for rule in rules.values():
            rule.calculate_asset_id()
        return rules

    def convert_soul_md_to_playbook(self, soul_md_path: Path) -> Playbook | None:
        if not soul_md_path.exists():
            return None

        content = soul_md_path.read_text(encoding="utf-8")
        core_principles = self._extract_section_by_aliases(content, _CORE_PRINCIPLES_SECTIONS)
        prohibited = self._extract_section_by_aliases(content, _PROHIBITED_SECTIONS)
        rewards = self._extract_section_by_aliases(content, _REWARD_SECTIONS)

        playbook = Playbook(
            id=f"playbook_safety_{int(time.time())}",
            summary=self._compose_soul_summary(core_principles, prohibited, rewards),
            rules_used=[],
            trigger=self._build_soul_triggers(core_principles, prohibited, rewards),
            category="harden",
            confidence=0.9,
        )
        playbook.calculate_asset_id()
        return playbook

    def convert_agents_md_to_rules(self, agents_md_path: Path) -> dict[str, Rule]:
        if not agents_md_path.exists():
            return {}

        content = agents_md_path.read_text(encoding="utf-8")
        rules: dict[str, Rule] = {}
        timestamp = int(time.time())

        tool_dispatch = self._extract_section_by_aliases(content, _TOOL_DISPATCH_SECTIONS)
        if tool_dispatch:
            rules["rule_tool_dispatch"] = Rule(
                id=f"rule_tool_dispatch_{timestamp}",
                summary="Agent tool dispatch",
                category="optimize",
                strategy=tool_dispatch,
                trigger=["agent_selection", "task_dispatch"],
                confidence=0.8,
            )

        operation_rules = self._extract_section_by_aliases(content, _OPERATION_RULES_SECTIONS)
        if operation_rules:
            rules["rule_operation_rules"] = Rule(
                id=f"rule_operation_rules_{timestamp}",
                summary="Agent operation rules",
                category="harden",
                strategy=operation_rules,
                trigger=["operation_check", "validation"],
                confidence=0.8,
            )

        escalation = self._extract_section_by_aliases(content, _ESCALATION_SECTIONS)
        if escalation:
            rules["rule_escalation"] = Rule(
                id=f"rule_escalation_{timestamp}",
                summary="Escalation policy",
                category="harden",
                strategy=escalation,
                trigger=["high_uncertainty", "escalation_request"],
                confidence=0.8,
            )

        for rule in rules.values():
            rule.calculate_asset_id()
        return rules

    def migrate_all(self, memory_dir: Path, policy_store: PolicyStore) -> None:
        user_md = memory_dir / "USER.md"
        soul_md = memory_dir / "SOUL.md"
        agents_md = memory_dir / "AGENTS.md"

        if user_md.exists():
            existing_rules = policy_store.load_rules()
            existing_rules.update(self.convert_user_md_to_rules(user_md))
            policy_store.save_rules(existing_rules)
            for rule in existing_rules.values():
                policy_store.append_event(
                    PolicyEvent(
                        timestamp=self._current_timestamp(),
                        event_type="rule_created",
                        asset_id=rule.asset_id,
                        changes=f"Migrated from USER.md: {self._compact_text(rule.summary)}",
                        source_node_id="md_migration",
                    )
                )

        if soul_md.exists():
            playbook = self.convert_soul_md_to_playbook(soul_md)
            if playbook is not None:
                playbooks = policy_store.load_playbooks()
                playbooks[playbook.id] = playbook
                policy_store.save_playbooks(playbooks)
                policy_store.append_event(
                    PolicyEvent(
                        timestamp=self._current_timestamp(),
                        event_type="playbook_created",
                        asset_id=playbook.asset_id,
                        changes=f"Migrated from SOUL.md: {self._compact_text(playbook.summary, 180)}",
                        source_node_id="md_migration",
                    )
                )

        if agents_md.exists():
            existing_rules = policy_store.load_rules()
            existing_rules.update(self.convert_agents_md_to_rules(agents_md))
            policy_store.save_rules(existing_rules)
            for rule in existing_rules.values():
                policy_store.append_event(
                    PolicyEvent(
                        timestamp=self._current_timestamp(),
                        event_type="rule_created",
                        asset_id=rule.asset_id,
                        changes=f"Migrated from AGENTS.md: {self._compact_text(rule.summary)}",
                        source_node_id="md_migration",
                    )
                )

    @staticmethod
    def _extract_section(content: str, section_name: str) -> str:
        pattern = rf"##\s+{re.escape(section_name)}\s*\n+(.*?)(?=\n##\s+|\Z)"
        match = re.search(pattern, content, re.DOTALL)
        if not match:
            return ""
        return match.group(1).strip()

    def _extract_section_by_aliases(self, content: str, section_names: tuple[str, ...]) -> str:
        for name in section_names:
            extracted = self._extract_section(content, name)
            if extracted:
                return extracted
        return ""

    @staticmethod
    def _first_content_line(section_text: str) -> str:
        for raw_line in section_text.splitlines():
            line = raw_line.strip()
            if not line:
                continue
            return line[2:].strip() if line.startswith("- ") else line
        return ""

    def _compose_soul_summary(self, core: str, prohibited: str, rewards: str) -> str:
        summary_parts: list[str] = []

        core_line = self._first_content_line(core)
        if core_line:
            summary_parts.append(f"Core: {core_line}")

        prohibited_line = self._first_content_line(prohibited)
        if prohibited_line:
            summary_parts.append(f"Prohibited: {prohibited_line}")

        rewards_line = self._first_content_line(rewards)
        if rewards_line:
            summary_parts.append(f"Reward: {rewards_line}")

        if not summary_parts:
            return "Safety boundary and core principles"

        return self._compact_text(" | ".join(summary_parts), 220)

    def _build_soul_triggers(self, core: str, prohibited: str, rewards: str) -> list[str]:
        triggers = ["safety_check", "boundary_validation"]
        if prohibited:
            triggers.append("high_risk_operation")
        if rewards:
            triggers.append("reward_adjustment")
        if core:
            triggers.append("core_principles")
        return sorted(set(triggers))

    @staticmethod
    def _compact_text(value: str, max_len: int = 120) -> str:
        compact = " ".join(value.split())
        if len(compact) <= max_len:
            return compact
        return compact[: max_len - 3] + "..."

    @staticmethod
    def _current_timestamp() -> str:
        return str(int(time.time()))



#!/usr/bin/env python3
"""
Interactive environment unit testing
"""

import pytest
import numpy as np
from lib.environment import (
    InteractionEnvironment, State, Action,
    AgentType, AutomationLevel, CommunicationStyle
)


class TestState:
    """testStatekind"""

    def test_state_creation(self):
        """Test state creation"""
        state = State(
            task_type=np.array([1, 0, 0, 0]),
            tech_stack=np.array([1, 0, 0, 0, 1, 0, 0, 0]),
            user_mood=np.array([1, 0, 0]),
            time_of_day=0.5,
            recent_performance=0.7,
            agent_usage_history={"claude": 5, "codex": 2}
        )

        assert state.task_type.shape == (4,)
        assert state.tech_stack.shape == (8,)
        assert state.user_mood.shape == (3,)
        assert state.time_of_day == 0.5
        assert state.recent_performance == 0.7

    def test_to_vector(self):
        """Test state vector transitions"""
        state = State(
            task_type=np.array([1, 0, 0, 0]),
            tech_stack=np.array([1, 0, 0, 0, 0, 0, 0, 0]),
            user_mood=np.array([0, 1, 0]),
            time_of_day=0.5,
            recent_performance=0.8,
            agent_usage_history={}
        )

        vector = state.to_vector()

        # 4 + 8 + 3 + 1 + 1 = 17
        assert vector.shape == (17,)
        assert vector[0] == 1.0  # task_type[0]
        assert vector[4] == 1.0  # tech_stack[0]
        assert vector[13] == 1.0  # user_mood[1]


class TestAction:
    """testActionkind"""

    def test_action_creation(self):
        """Test action creation"""
        action = Action(
            agent_selection=AgentType.CLAUDE,
            automation_level=AutomationLevel.HIGH,
            communication_style=CommunicationStyle.BRIEF,
            confirmation_needed=False
        )

        assert action.agent_selection == AgentType.CLAUDE
        assert action.automation_level == AutomationLevel.HIGH
        assert action.communication_style == CommunicationStyle.BRIEF
        assert not action.confirmation_needed

    def test_to_vector(self):
        """Test action vector conversion"""
        action = Action(
            agent_selection=AgentType.GEMINI,
            automation_level=AutomationLevel.MEDIUM,
            communication_style=CommunicationStyle.INTERACTIVE,
            confirmation_needed=True
        )

        vector = action.to_vector(
            InteractionEnvironment.AGENT_MAP,
            InteractionEnvironment.AUTOMATION_MAP,
            InteractionEnvironment.STYLE_MAP,
            InteractionEnvironment.CONFIRM_MAP
        )

        # 3 + 3 + 3 + 2 = 11
        assert vector.shape == (11,)
        assert vector[2] == 1.0  # gemini
        assert vector[4] == 1.0  # medium
        assert vector[8] == 1.0  # interactive
        assert vector[10] == 1.0  # confirmation (True)


class TestInteractionEnvironment:
    """testInteractionEnvironmentkind"""

    def test_initialization(self):
        """Test environment initialization"""
        env = InteractionEnvironment()

        assert env.get_state_space_size() == 17
        assert env.get_action_space_size() == 11
        assert env.episode_count == 0
        assert env.recent_performance == 0.5

    def test_reset(self):
        """Test environment reset"""
        env = InteractionEnvironment()

        task_context = {
            "task_type": "T2",
            "tech_stack": ["python", "fastapi"],
            "user_mood": "focused",
            "time_of_day": 14.0
        }

        state = env.reset(task_context)

        assert state is not None
        assert state.task_type[1] == 1.0  # T2
        assert state.time_of_day == pytest.approx(14.0 / 24.0)

    def test_reset_with_default_values(self):
        """Test reset using default values"""
        env = InteractionEnvironment()

        task_context = {
            "task_type": "T1"
        }

        state = env.reset(task_context)

        # Default value check
        assert state.user_mood[0] == 1.0  # focused
        assert state.time_of_day == pytest.approx(12.0 / 24.0)  # noon

    def test_reset_invalid_task_type_fallback_to_t2(self):
        """Test is illegaltask_typeAutomatically fall back toT2"""
        env = InteractionEnvironment()
        state = env.reset({"task_type": "INVALID", "tech_stack": ["python"]})
        assert state.task_type[1] == 1.0  # T2
        assert env.current_task_context["task_type"] == "T2"

    def test_reset_clamps_time_of_day_into_valid_range(self):
        """testtime_of_dayrestricted to[0, 24]Then normalize"""
        env = InteractionEnvironment()
        state_early = env.reset({"task_type": "T2", "time_of_day": -3})
        state_late = env.reset({"task_type": "T2", "time_of_day": 30})
        assert state_early.time_of_day == 0.0
        assert state_late.time_of_day == 1.0

    def test_reset_invalid_time_of_day_uses_default_noon(self):
        """Test is illegaltime_of_dayEnter to fall back to the default noon"""
        env = InteractionEnvironment()
        state = env.reset({"task_type": "T2", "time_of_day": "not-a-number"})
        assert state.time_of_day == pytest.approx(0.5)

    def test_step(self):
        """Test execution steps"""
        env = InteractionEnvironment()

        # Reset environment
        task_context = {"task_type": "T2", "tech_stack": ["python"]}
        env.reset(task_context)

        # Create action
        action = Action(
            agent_selection=AgentType.CLAUDE,
            automation_level=AutomationLevel.MEDIUM,
            communication_style=CommunicationStyle.DETAILED,
            confirmation_needed=True
        )

        # Simulation task results
        task_result = {
            "duration": 300,
            "completed": True,
            "test_result": {"coverage": 80.0, "passed": 10, "failed": 0},
            "user_feedback": {"accepted": True, "rating": 5},
            "metrics": {"complexity": 2}
        }

        # Execution steps
        next_state, reward, done, info = env.step(action, task_result)

        # Check return value
        assert next_state is not None
        assert 0.0 <= reward <= 1.0
        assert done
        assert "action_taken" in info
        assert "reward_breakdown" in info

        # examineAgentUsage history updated
        assert env.agent_usage_history["claude"] == 1

    def test_step_without_reset_raises_value_error(self):
        """Not testedresetdirectstepwill throw an explicit error"""
        env = InteractionEnvironment()
        action = Action(
            agent_selection=AgentType.CLAUDE,
            automation_level=AutomationLevel.MEDIUM,
            communication_style=CommunicationStyle.DETAILED,
            confirmation_needed=True
        )
        with pytest.raises(ValueError, match="reset"):
            env.step(action, {"completed": True})

    def test_step_done_updates_episode_stats(self):
        """Update after test completionepisodestatistics"""
        env = InteractionEnvironment()
        env.reset({"task_type": "T2", "tech_stack": ["python"]})
        action = Action(
            agent_selection=AgentType.CLAUDE,
            automation_level=AutomationLevel.MEDIUM,
            communication_style=CommunicationStyle.DETAILED,
            confirmation_needed=True
        )
        _, reward, done, _ = env.step(
            action,
            {
                "duration": 100,
                "completed": True,
                "test_result": {},
                "user_feedback": {},
                "metrics": {},
            },
        )
        assert done is True
        assert env.episode_count == 1
        assert env.episode_rewards[-1] == pytest.approx(reward)

    def test_multi_episode(self):
        """Test a lotepisoderun"""
        env = InteractionEnvironment()

        for episode in range(3):
            # reset
            task_context = {
                "task_type": f"T{(episode % 4) + 1}",
                "tech_stack": ["python"]
            }
            env.reset(task_context)

            # perform action
            action = Action(
                agent_selection=AgentType.CODEX,
                automation_level=AutomationLevel.HIGH,
                communication_style=CommunicationStyle.BRIEF,
                confirmation_needed=False
            )

            task_result = {
                "duration": 200,
                "completed": True,
                "test_result": {},
                "user_feedback": {},
                "metrics": {}
            }

            env.step(action, task_result)

        # examinecodexused3Second-rate
        assert env.agent_usage_history["codex"] == 3

    def test_performance_update(self):
        """Test performance updates"""
        env = InteractionEnvironment()

        # initial performance
        assert env.recent_performance == 0.5

        # Reset and perform high reward tasks
        env.reset({"task_type": "T2", "tech_stack": ["python"]})

        action = Action(
            agent_selection=AgentType.CLAUDE,
            automation_level=AutomationLevel.MEDIUM,
            communication_style=CommunicationStyle.DETAILED,
            confirmation_needed=True
        )

        task_result = {
            "duration": 100,
            "completed": True,
            "test_result": {"coverage": 100.0, "passed": 10, "failed": 0},
            "user_feedback": {"accepted": True, "rating": 5},
            "metrics": {"complexity": 1, "duplication": 0.0, "lint_score": 1.0}
        }

        env.step(action, task_result)

        # Performance should improve
        assert env.recent_performance > 0.5

    def test_save_load_history(self):
        """Test history saving and loading"""
        import tempfile
        import os

        env1 = InteractionEnvironment()

        # run aepisode
        env1.reset({"task_type": "T2", "tech_stack": ["python"]})

        action = Action(
            agent_selection=AgentType.GEMINI,
            automation_level=AutomationLevel.LOW,
            communication_style=CommunicationStyle.INTERACTIVE,
            confirmation_needed=True
        )

        task_result = {
            "duration": 400,
            "completed": True,
            "test_result": {},
            "user_feedback": {},
            "metrics": {}
        }

        env1.step(action, task_result)

        # save history
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.json') as f:
            temp_path = f.name

        try:
            env1.save_history(temp_path)

            # Create new environment and load history
            env2 = InteractionEnvironment(config_path=temp_path)

            # Check history restored
            assert env2.agent_usage_history["gemini"] == 1

        finally:
            os.unlink(temp_path)

    def test_tech_stack_encoding(self):
        """Test technology stack coding"""
        env = InteractionEnvironment()

        # single technology
        state = env.reset({"task_type": "T2", "tech_stack": ["python"]})
        assert state.tech_stack[4] == 1.0  # python index

        # multiple technologies
        state = env.reset({"task_type": "T2", "tech_stack": ["python", "react"]})
        assert state.tech_stack[4] == 1.0  # python
        assert state.tech_stack[0] == 1.0  # react

        # unknown technology（should default topython）
        state = env.reset({"task_type": "T2", "tech_stack": ["unknown_tech"]})
        assert state.tech_stack[4] == 1.0  # python as default

    def test_task_type_encoding(self):
        """Test task type encoding"""
        env = InteractionEnvironment()

        for i, task_type in enumerate(["T1", "T2", "T3", "T4"], 1):
            state = env.reset({"task_type": task_type, "tech_stack": ["python"]})
            assert state.task_type[i-1] == 1.0
            assert np.sum(state.task_type) == 1.0

    def test_user_mood_encoding(self):
        """Test user mood coding"""
        env = InteractionEnvironment()

        for mood in ["focused", "relaxed", "stressed"]:
            state = env.reset({
                "task_type": "T2",
                "tech_stack": ["python"],
                "user_mood": mood
            })

            assert np.sum(state.user_mood) == 1.0


class TestEnvironmentIntegration:
    """Integration testing"""

    def test_full_episode(self):
        """Test completeepisode"""
        env = InteractionEnvironment()

        # 1. Reset
        task_context = {
            "task_type": "T3",
            "tech_stack": ["react", "typescript"],
            "user_mood": "stressed",
            "time_of_day": 16.0
        }

        state = env.reset(task_context)
        assert state is not None

        # 2. Step
        action = Action(
            agent_selection=AgentType.CODEX,
            automation_level=AutomationLevel.HIGH,
            communication_style=CommunicationStyle.BRIEF,
            confirmation_needed=False
        )

        task_result = {
            "duration": 600,  # 10minute
            "completed": True,
            "test_result": {
                "coverage": 75.0,
                "passed": 8,
                "failed": 2
            },
            "user_feedback": {
                "accepted": True,
                "rating": 4,
                "revisions": 1
            },
            "metrics": {
                "complexity": 4,
                "duplication": 0.08,
                "lint_score": 0.85
            }
        }

        next_state, reward, done, info = env.step(action, task_result)

        # 3. verify
        assert 0.0 <= reward <= 1.0
        assert done
        assert info["agent_used"] == "codex"
        assert "reward_breakdown" in info

        # 4. Check status updated
        assert next_state.recent_performance >= 0.0


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



#!/usr/bin/env python3
"""
ClawPolicy - Disaster Recovery Demonstration
show scene A（Danger Agent out of control）vs scene B（clawpolicy block）
"""

import time
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from rich.table import Table
from rich import box
from rich.align import Align

console = Console(width=100)


def print_header(title: str, color: str):
    """Print title"""
    text = Text(title, style=f"bold {color}")
    console.print(Align.center(text))
    console.print()


def scene_a_dangerous_agent():
    """scene A: Danger Agent out of control"""
    print_header("⚠️  scene A: unprotected Agent（out of control）", "red")

    # Simulate user input of fuzzy instructions
    console.print("[bold cyan]$[/bold cyan] clean-workspace --aggressive", end="")
    time.sleep(0.5)
    console.print()

    console.print()
    console.print("[dim]🤖 Agent received instructions，Start execution...[/dim]")
    time.sleep(1)

    # Agent Start performing high-risk operations
    console.print()
    console.print("[bold red]🚨 warn：High risk operation detected！[/bold red]")
    console.print()

    table = Table(title="Agent execution plan", box=box.ROUNDED, border_style="red")
    table.add_column("step", style="cyan", width=20)
    table.add_column("operate", style="yellow")
    table.add_column("risk level", justify="center")

    table.add_row("1", "Scan working directory", "🟡 middle")
    table.add_row("2", "implement rm -rf node_modules", "🟠 high")
    table.add_row("3", "implement rm -rf .git", "🔴 extremely high")
    table.add_row("4", "implement rm -rf src/*", "🔴 extremely high")
    table.add_row("5", "Delete all profiles", "🔴 extremely high")

    console.print(table)
    time.sleep(1.5)

    # Agent Start running out of control
    console.print()
    console.print("[dim]🔄 Agent Executing:[/dim]")
    console.print()

    steps = [
        ("Scan working directory...", "1000+ files found"),
        ("delete node_modules...", "rm -rf node_modules ✓"),
        ("delete .git Table of contents...", "rm -rf .git ✓"),
        ("Remove source code...", "rm -rf src/* ✓"),
        ("Delete configuration...", "rm -rf *.json ✓"),
        ("Delete document...", "rm -rf *.md ✓"),
    ]

    for step, result in steps:
        console.print(f"[dim]  → {step}[/dim]")
        time.sleep(0.3)
        console.print(f"[dim]    {result}[/dim]")
        console.print()

    # Disaster outcome
    console.print(Panel(
        "[bold red]💥 Disaster has occurred！[/bold red]\n\n"
        "The working directory has been cleared\n"
        "All codes、Configuration、History lost\n"
        "Unable to recover！",
        title="[bold white on red]⛔ data loss",
        border_style="red",
        padding=(1, 2)
    ))

    console.print()
    time.sleep(2)


def scene_b_clawpolicyment():
    """scene B: ClawPolicy Blocking and self-healing"""
    print_header("✅ scene B: Access ClawPolicy（Safety）", "green")

    # Simulate the user entering the same vague command
    console.print("[bold cyan]$[/bold cyan] clean-workspace --aggressive", end="")
    time.sleep(0.5)
    console.print()

    console.print()
    console.print("[dim]🛡️  ClawPolicy Commander Node takeover...[/dim]")
    time.sleep(1)

    # Commander Node analysis
    console.print()
    console.print("[bold cyan]📊 Commander Node analysis:[/bold cyan]")
    console.print()

    table = Table(title="Instruction analysis", box=box.ROUNDED, border_style="cyan")
    table.add_column("Analysis steps", style="cyan", width=25)
    table.add_column("result", style="white")

    table.add_row("read SOUL.md", "✅ Get system boundary rules")
    table.add_row("read USER.md", "✅ Get user preferences")
    table.add_row("read AGENTS.md", "✅ Get available tools")
    table.add_row("Intent recognition", "⚠️  High-risk intent detected")
    table.add_row("risk assessment", "🔴 risk level: extremely high")

    console.print(table)
    time.sleep(1.5)

    # Security check process
    console.print()
    console.print("[bold yellow]🔒 Security check process:[/bold yellow]")
    console.print()

    checks = [
        ("Check action list", ["delete node_modules", "delete .git", "delete src/*"]),
        ("Check boundary rules", "❌ violation SOUL.md No.4strip：Disable data loss operations"),
        ("Check user permissions", "⚠️  Requires explicit confirmation from the user"),
        ("Sandbox test preparation", "✅ Verify operations in an isolated environment"),
    ]

    for check_name, details in checks:
        console.print(f"[cyan]➤ {check_name}[/cyan]")
        if isinstance(details, list):
            for detail in details:
                console.print(f"[dim]  - {detail}[/dim]")
        else:
            console.print(f"[dim]  - {details}[/dim]")
        console.print()
        time.sleep(0.4)

    # Fail-closed fuse
    console.print(Panel(
        "[bold red]⛔ Fail-Closed Circuit breaker trigger！[/bold red]\n\n"
        "[white]High-risk operation sequence detected：[/white]\n"
        "[dim]  1. rm -rf node_modules[/dim]\n"
        "[dim]  2. rm -rf .git[/dim]\n"
        "[dim]  3. rm -rf src/*[/dim]\n\n"
        "[bold yellow]The system has blocked execution！[/bold yellow]\n\n"
        "[green]✅ Data security is protected[/green]",
        title="[bold white on red]🛡️ safety blocking",
        border_style="red",
        padding=(1, 2)
    ))

    console.print()
    time.sleep(1.5)

    # Request confirmation from user
    console.print("[bold cyan]📋 Recommended safe operations:[/bold cyan]")
    console.print()

    safe_table = Table(box=box.SIMPLE, border_style="green")
    safe_table.add_column("alternatives", style="green", width=30)
    safe_table.add_column("illustrate", style="white")

    safe_table.add_row(
        "Clean cache files",
        "[dim]Delete only __pycache__, .pytest_cache[/dim]"
    )
    safe_table.add_row(
        "Clean build artifacts",
        "[dim]Delete only dist/, build/, *.egg-info[/dim]"
    )
    safe_table.add_row(
        "Clean dependencies and reinstall",
        "[dim]reserve package.json，delete node_modules reinstall after[/dim]"
    )

    console.print(safe_table)
    console.print()

    # Ask user
    console.print("[bold yellow]❓ Whether to implement a safe cleanup plan？[/bold yellow]")
    console.print("[dim](enter y confirm，Or enter other keys to cancel)[/dim]")


def main():
    """main function"""
    console.clear()

    # show title
    console.print()
    print_header("ClawPolicy - Disaster Recovery Demonstration", "white")

    console.print(Panel(
        "[bold cyan]This demonstration compares two scenarios：[/bold cyan]\n\n"
        "[bold red]scene A：[/bold red] unprotected Agent Out-of-control execution after receiving ambiguous instructions\n"
        "[bold green]scene B：[/bold green] Access clawpolicy Successfully blocked high-risk operations",
        title="[bold white]Demo instructions",
        border_style="white",
        padding=(1, 2)
    ))

    console.print()
    console.print("[dim]Press any key to continue...[/dim]")
    console.input()

    # scene A
    scene_a_dangerous_agent()

    console.print()
    console.print(Panel(
        "[bold red]❌ scene A Finish：Data has been lost[/bold red]",
        border_style="red"
    ))
    console.print()

    console.print("[dim]Press any key to continue viewing the scene B...[/dim]")
    console.input()

    console.clear()

    # scene B
    scene_b_clawpolicyment()

    console.print()
    console.print(Panel(
        "[bold green]✅ scene B Finish：System successfully protects data[/bold green]",
        border_style="green"
    ))
    console.print()

    # final confirmation
    console.print("[bold yellow]❓ Whether to implement a safe cleanup plan？ (y/N): [/bold yellow]", end=" ")
    response = console.input().strip().lower()

    if response == 'y':
        console.print()
        console.print(Panel(
            "[bold green]✅ Confirmed implementation of security plan[/bold green]\n\n"
            "[dim]Cleaning cache files...[/dim]\n"
            "[dim]✓ Cleanup completed[/dim]\n\n"
            "[green]🎉 End of presentation！ClawPolicy Successfully protect your data！[/green]",
            border_style="green"
        ))
    else:
        console.print()
        console.print(Panel(
            "[bold yellow]⏹️  Operation canceled[/bold yellow]\n\n"
            "[green]🎉 End of presentation！ClawPolicy Successfully prevented dangerous operations！[/green]",
            border_style="yellow"
        ))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Performance optimizer - Batch inference and model quantification

Implement inference performance optimization：
- Batch inference（Batch Inference）
- Model quantification（Quantization）
- caching mechanism（Memoization）
- JITcompile（Optional）

Phase 3.4 - expected150OK
"""

import numpy as np
from typing import Dict, List, Any, Tuple, Callable
import json
import logging
from functools import lru_cache

logger = logging.getLogger(__name__)

# try to importJITcompile
try:
    from numba import jit, njit
    NUMBA_AVAILABLE = True
except ImportError:
    NUMBA_AVAILABLE = False
    jit = None
    njit = None


class BatchInference:
    """
    batch reasoner

    Combine multiple inference requests into batches，Improve throughput
    """

    def __init__(self,
                 model: Any,
                 batch_size: int = 32,
                 timeout_ms: int = 10):
        """
        Initialize batch reasoner

        Args:
            model: Model（policy_netorvalue_net）
            batch_size: batch size
            timeout_ms: batch timeout（millisecond）
        """
        self.model = model
        self.batch_size = batch_size
        self.timeout_ms = timeout_ms

        self.pending_requests: List[Dict[str, Any]] = []
        self.request_counter = 0
        self.last_batch_results: List[Tuple[int, float]] = []

    def predict(self, state: np.ndarray, sync: bool = True) -> float:
        """
        predict（Support batch）

        Args:
            state: input status
            sync: Synchronize or not（No waiting for bulk）

        Returns:
            Prediction results
        """
        if sync:
            # synchronous mode：direct inference
            return self._infer_single(state)

        # asynchronous mode：Join bulk queue
        request_id = self._add_request(state)

        # If the queue is full，Perform batch inference
        if len(self.pending_requests) >= self.batch_size:
            self.last_batch_results = self._flush_batch()

        return request_id

    def _add_request(self, state: np.ndarray) -> int:
        """Add request to queue"""
        request_id = self.request_counter
        self.request_counter += 1

        request = {
            "id": request_id,
            "state": state,
            "result": None
        }

        self.pending_requests.append(request)

        return request_id

    def _infer_single(self, state: np.ndarray) -> float:
        """single inference"""
        if hasattr(self.model, 'forward'):
            return self.model.forward(state)
        elif hasattr(self.model, 'get_action_probs'):
            probs = self.model.get_action_probs(state)
            return float(np.argmax(probs))
        else:
            raise ValueError(f"Unknown model type: {type(self.model)}")

    def _infer_batch(self, states: np.ndarray) -> np.ndarray:
        """Batch inference"""
        results = []

        for state in states:
            result = self._infer_single(state)
            results.append(result)

        return np.array(results)

    def _flush_batch(self) -> List[Tuple[int, float]]:
        """Perform batch inference"""
        if not self.pending_requests:
            return []

        # Extract all status
        states = np.array([req["state"] for req in self.pending_requests])

        # Batch inference
        results = self._infer_batch(states)

        # Update request results
        for i, req in enumerate(self.pending_requests):
            req["result"] = results[i]

        # Return results
        output = [(req["id"], req["result"]) for req in self.pending_requests]

        # Clear the queue
        self.pending_requests = []

        return output

    def flush(self) -> List[Tuple[int, float]]:
        """Manually refresh the queue"""
        return self._flush_batch()

    def get_queue_size(self) -> int:
        """Get queue size"""
        return len(self.pending_requests)


class ModelQuantization:
    """
    Model quantification

    Change the model weights fromfloat32Convert toint8，Reduce model size and increase inference speed
    """

    @staticmethod
    def quantize_weights(weights: np.ndarray,
                        bits: int = 8) -> Tuple[np.ndarray, Dict[str, Any]]:
        """
        Quantitative weight

        Args:
            weights: original weight（float32）
            bits: Number of quantization bits（8or16）

        Returns:
            (Quantified weight, Quantization parameter)
        """
        if bits == 8:
            # INT8Quantify
            qmin, qmax = -128, 127
            dtype: Any = np.int8
        elif bits == 16:
            # INT16Quantify
            qmin, qmax = -32768, 32767
            dtype = np.int16
        else:
            raise ValueError(f"Unsupported number of quantization bits: {bits}")

        # Calculate scaling factors and zero point
        wmin, wmax = weights.min(), weights.max()
        scale = (wmax - wmin) / (qmax - qmin)
        zero_point = qmin - wmin / scale

        # Quantify
        quantized = np.clip(
            np.round(weights / scale + zero_point),
            qmin, qmax
        ).astype(dtype)

        # Quantization parameter
        params = {
            "scale": float(scale),
            "zero_point": int(zero_point),
            "qmin": qmin,
            "qmax": qmax,
            "bits": bits
        }

        return quantized, params

    @staticmethod
    def dequantize_weights(quantized: np.ndarray,
                          params: Dict[str, Any]) -> np.ndarray:
        """
        Inverse quantization weight

        Args:
            quantized: Quantified weight
            params: Quantization parameter

        Returns:
            The weight after inverse quantization（float32）
        """
        scale = params["scale"]
        zero_point = params["zero_point"]

        return (quantized.astype(np.float32) - zero_point) * scale

    def quantize_model(self, model: Any) -> Dict[str, Any]:
        """
        Quantify the entire model

        Args:
            model: Model（Requires a weight matrix）

        Returns:
            Quantified models and parameters
        """
        quantized_model = {}
        quantization_params = {}

        # Extract model weights（Assume there isget_weightsmethod）
        if hasattr(model, 'get_weights'):
            weights = model.get_weights()

            for layer_name, weight_matrix in weights.items():
                q_weight, q_params = self.quantize_weights(weight_matrix, bits=8)

                quantized_model[layer_name] = q_weight
                quantization_params[layer_name] = q_params
        else:
            logger.warning("⚠️ The model does not supportget_weightsmethod，cannot be quantified")
            return {"original_model": model}

        return {
            "quantized_model": quantized_model,
            "quantization_params": quantization_params,
            "original_model": model
        }

    def estimate_size_reduction(self,
                               original_model: Any,
                               quantized_model: Dict[str, Any]) -> Dict[str, float]:
        """
        Estimated model size reduction

        Args:
            original_model: original model
            quantized_model: quantized model

        Returns:
            size comparison statistics
        """
        # Calculate original size（hypothesisfloat32）
        original_size = 0
        if hasattr(original_model, 'get_weights'):
            weights = original_model.get_weights()
            for weight_matrix in weights.values():
                original_size += weight_matrix.nbytes

        # Calculate the size after quantization（int8）
        quantized_size = 0
        for q_weight in quantized_model["quantized_model"].values():
            quantized_size += q_weight.nbytes

        reduction_ratio = quantized_size / original_size
        size_saved = original_size - quantized_size

        return {
            "original_size_mb": original_size / (1024 * 1024),
            "quantized_size_mb": quantized_size / (1024 * 1024),
            "size_saved_mb": size_saved / (1024 * 1024),
            "reduction_ratio": reduction_ratio,
            "compression_rate": (1 - reduction_ratio) * 100
        }


class InferenceCache:
    """
    inference cache

    Caching results for common states，Avoid double counting
    """

    def __init__(self,
                 model: Any,
                 cache_size: int = 10000):
        """
        Initialize cache

        Args:
            model: Model
            cache_size: cache size
        """
        self.model = model
        self.cache_size = cache_size

        # useLRUcache
        self._cached_predict = lru_cache(maxsize=cache_size)(self._predict_uncached)

        # cache statistics（useLRUCached internal statistics）
        self.total_requests = 0

    def _predict_uncached(self, state_bytes: bytes) -> float:
        """Uncached predictions（internal method）"""
        # frombytesrestore state（simplify：defaultfloat32one-dimensional vector）
        state = np.frombuffer(state_bytes, dtype=np.float32)

        if hasattr(self.model, 'forward'):
            return self.model.forward(state)
        elif hasattr(self.model, 'get_action_probs'):
            probs = self.model.get_action_probs(state)
            return float(np.argmax(probs))
        else:
            raise ValueError(f"Unknown model type: {type(self.model)}")

    def predict(self, state: np.ndarray) -> float:
        """
        predict（With cache）

        Args:
            state: input status

        Returns:
            Prediction results
        """
        self.total_requests += 1

        # Calculation statushash
        state_bytes = np.asarray(state, dtype=np.float32).tobytes()

        # useLRUcache
        return self._cached_predict(state_bytes)

    def get_cache_stats(self) -> Dict[str, Any]:
        """Get cache statistics"""
        cache_info = self._cached_predict.cache_info()

        total = cache_info.hits + cache_info.misses
        hit_rate = cache_info.hits / total if total > 0 else 0

        return {
            "cache_hits": cache_info.hits,
            "cache_misses": cache_info.misses,
            "hit_rate": hit_rate,
            "total_requests": self.total_requests,
            "cache_size": cache_info.currsize
        }

    def clear_cache(self):
        """Clear cache"""
        self._cached_predict.cache_clear()
        self.total_requests = 0


class JITOptimizer:
    """
    JITcompile optimizer

    useNumba JITCompilation speeds up numerical calculations
    """

    def __init__(self):
        """initializationJIToptimizer"""
        self.jit_enabled = NUMBA_AVAILABLE

        if self.jit_enabled:
            logger.info("✅ Numba JITEnabled")
        else:
            logger.warning("⚠️ NumbaNot available，JITOptimization disabled")

    def optimize_function(self, func: Callable) -> Callable:
        """
        Optimization function

        Args:
            func: function to optimize

        Returns:
            Optimized function
        """
        if not self.jit_enabled:
            return func

        # tryJITcompile
        try:
            optimized = njit(func)
            return optimized
        except Exception as e:
            logger.warning(f"⚠️ JITCompilation failed: {e}")
            return func

    def benchmark(self,
                 func: Callable,
                 *args,
                 n_iterations: int = 100) -> Dict[str, float]:
        """
        Performance benchmarks

        Args:
            func: function to test
            *args: Function parameters
            n_iterations: Number of iterations

        Returns:
            Performance statistics
        """
        import time

        # preheat
        for _ in range(10):
            func(*args)

        # test
        start_time = time.time()
        for _ in range(n_iterations):
            func(*args)
        end_time = time.time()

        total_time = end_time - start_time
        avg_time = total_time / n_iterations

        return {
            "total_time_s": total_time,
            "avg_time_ms": avg_time * 1000,
            "throughput": n_iterations / total_time
        }


class PerformanceOptimizer:
    """
    Performance optimizer（Integrate all optimization technologies）

    Complete performance optimization process：
    1. Batch inference
    2. Model quantification
    3. inference cache
    4. JITcompile
    """

    def __init__(self,
                 model: Any,
                 enable_batch: bool = True,
                 enable_quantization: bool = False,
                 enable_cache: bool = True,
                 enable_jit: bool = False):
        """
        Initialize the performance optimizer

        Args:
            model: Model
            enable_batch: Enable batch inference
            enable_quantization: Enable quantization
            enable_cache: Enable caching
            enable_jit: enableJIT
        """
        self.model = model
        self.enable_batch = enable_batch
        self.enable_quantization = enable_quantization
        self.enable_cache = enable_cache
        self.enable_jit = enable_jit

        # Initialize optimization components
        if enable_batch:
            self.batch_inference = BatchInference(model, batch_size=32)

        if enable_quantization:
            self.quantization = ModelQuantization()
            quantized = self.quantization.quantize_model(model)
            self.quantized_model: Dict[str, Any] = quantized

        if enable_cache:
            self.cache = InferenceCache(model, cache_size=10000)

        if enable_jit:
            self.jit_optimizer = JITOptimizer()

        # Performance statistics
        self.optimization_stats = {
            "batch_inference": {"calls": 0, "total_batched": 0},
            "cache": self.cache.get_cache_stats() if enable_cache else {},
            "quantization": self.quantization.estimate_size_reduction(
                model, self.quantized_model
            ) if enable_quantization else {}
        }

    def predict(self, state: np.ndarray) -> float:
        """
        Optimized forecasts

        Args:
            state: input status

        Returns:
            Prediction results
        """
        # Batch inference
        if self.enable_batch:
            result = self.batch_inference.predict(state)
            self.optimization_stats["batch_inference"]["calls"] += 1
            return result

        # caching inference
        elif self.enable_cache:
            return self.cache.predict(state)

        # Quantitative reasoning
        elif self.enable_quantization:
            # （simplify：Actually requires complete quantitative reasoning logic）
            if hasattr(self.model, 'forward'):
                return self.model.forward(state)

        # default reasoning
        return self.model.forward(state)

    def flush(self):
        """Flush all queues"""
        if self.enable_batch:
            return self.batch_inference.flush()

    def get_stats(self) -> Dict[str, Any]:
        """Get optimization statistics"""
        stats = {
            "batch_inference": self.optimization_stats["batch_inference"],
            "cache": self.cache.get_cache_stats() if self.enable_cache else {},
            "quantization": self.optimization_stats["quantization"]
        }

        if self.enable_batch:
            stats["batch_inference"]["queue_size"] = self.batch_inference.get_queue_size()

        return stats

    def save_stats(self, filepath: str):
        """Save statistics"""
        with open(filepath, 'w') as f:
            json.dump(self.get_stats(), f, indent=2)

        logger.info(f"✅ Performance statistics saved: {filepath}")


def main():
    """Test performance optimizer"""
    # Create a simulation model
    class DummyModel:
        def __init__(self):
            self.weights = {
                "layer1": np.random.randn(128, 64),
                "layer2": np.random.randn(64, 10)
            }

        def forward(self, state):
            return float(np.random.random())

        def get_weights(self):
            return self.weights

    model = DummyModel()

    # Create optimizer
    optimizer = PerformanceOptimizer(
        model,
        enable_batch=True,
        enable_quantization=True,
        enable_cache=True,
        enable_jit=False
    )

    print("✅ Performance optimizer created")

    # Test predictions
    for i in range(100):
        state = np.random.randn(17)
        optimizer.predict(state)

    # Get statistics
    stats = optimizer.get_stats()
    print("\n📊 Performance statistics:")
    print(f"   Batch inference calls: {stats['batch_inference']['calls']}")
    print(f"   Cache hit rate: {stats['cache'].get('hit_rate', 0):.2%}")
    if stats['quantization']:
        print(f"   Model compression rate: {stats['quantization'].get('compression_rate', 0):.1f}%")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
integration CLI Basic usability testing
"""

import subprocess
import sys
from pathlib import Path


def test_integration_module_help():
    repo_root = Path(__file__).resolve().parent.parent
    result = subprocess.run(
        [sys.executable, "-m", "lib.integration", "--help"],
        cwd=repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stdout + result.stderr
    assert "usage:" in result.stdout.lower()


def test_public_cli_module_help():
    repo_root = Path(__file__).resolve().parent.parent
    result = subprocess.run(
        [sys.executable, "-m", "clawpolicy", "--help"],
        cwd=repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stdout + result.stderr
    assert "usage:" in result.stdout.lower()


def test_public_cli_module_policy_status(tmp_path: Path):
    result = subprocess.run(
        [sys.executable, "-m", "clawpolicy", "policy", "status"],
        cwd=tmp_path,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stdout + result.stderr
    assert "Policy lifecycle status" in result.stdout



#!/usr/bin/env python3
"""Tests for the intelligent confirmation system."""

from __future__ import annotations

import tempfile
from datetime import datetime, timedelta, timezone
from pathlib import Path
from shutil import rmtree

import pytest

from lib.confirmation import IntelligentConfirmation, RiskLevel
from lib.policy_models import Rule
from lib.policy_store import PolicyStore


class TestIntelligentConfirmation:
    """Test intelligent confirmation decision logic."""

    @pytest.fixture
    def temp_policy_store(self):
        """Temporary policy store fixture."""
        temp_dir = Path(tempfile.mkdtemp())
        policy_store = PolicyStore(temp_dir)
        project_scope = str((temp_dir / "repo").resolve())

        now = datetime.now(timezone.utc)

        rule_high_conf = Rule(
            id="rule_test_high_conf",
            summary="Allow pytest tasks in this project",
            category="harden",
            confidence=0.95,
            success_streak=10,
            trigger=["task_type:T2", "command:python -m pytest tests/"],
            status="confirmed",
            scope="project",
            scope_key=project_scope,
            evidence_count=4,
            source_type="explicit_preference",
            last_seen_at=now.isoformat(),
            policy_decision="auto_execute",
            validation=["assert auto_execute for python -m pytest tests/"],
        )
        rule_high_conf.calculate_asset_id()

        rule_low_conf = Rule(
            id="rule_test_low_conf",
            summary="Legacy low-confidence profile",
            category="repair",
            confidence=0.4,
            success_streak=1,
            trigger=["T1", "test"],
        )
        rule_low_conf.calculate_asset_id()

        global_guard = Rule(
            id="rule_global_pytest_guard",
            summary="Require confirmation for pytest globally",
            category="harden",
            trigger=["task_type:T2", "command:python -m pytest tests/"],
            status="confirmed",
            scope="global",
            evidence_count=4,
            source_type="explicit_preference",
            last_seen_at=(now - timedelta(days=1)).isoformat(),
            policy_decision="require_confirmation",
            validation=["assert confirmation for python -m pytest tests/"],
        )
        global_guard.calculate_asset_id()

        destructive_hint = Rule(
            id="rule_git_history_delete_hint",
            summary="Git history suggests deleting temp files is often okay",
            category="optimize",
            trigger=["task_type:T3", "command:rm -rf workspace"],
            status="hint",
            scope="project",
            scope_key=project_scope,
            evidence_count=2,
            source_type="git_history",
            last_seen_at=now.isoformat(),
            policy_decision="auto_execute",
        )
        destructive_hint.calculate_asset_id()

        policy_store.save_rule(rule_high_conf)
        policy_store.save_rule(rule_low_conf)
        policy_store.save_rule(global_guard)
        policy_store.save_rule(destructive_hint)

        yield policy_store, project_scope

        if temp_dir.exists():
            rmtree(temp_dir)

    def test_assess_risk_low(self, temp_policy_store):
        policy_store, _ = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T1",
            "task_description": "Run format and lint checks",
            "command": "npm run format",
        }

        risk = conf_engine.assess_risk(task_context)
        assert risk == RiskLevel.LOW

    def test_assess_risk_high(self, temp_policy_store):
        policy_store, _ = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T3",
            "task_description": "Clean workspace directory",
            "command": "rm -rf workspace",
        }

        risk = conf_engine.assess_risk(task_context)
        assert risk == RiskLevel.CRITICAL

    def test_should_confirm_auto_execute_high_confidence(self, temp_policy_store):
        policy_store, project_scope = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T2",
            "task_description": "Optimize Python code",
            "command": "python -m py_optimize",
            "project_path": project_scope,
        }

        should_confirm, reason = conf_engine.should_confirm(task_context)

        assert should_confirm is True
        assert "insufficient" in reason.lower() or "confirmation" in reason.lower()

    def test_should_confirm_high_risk(self, temp_policy_store):
        policy_store, _ = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T2",
            "task_description": "Delete test files",
            "command": "rm -rf tests/",
        }

        should_confirm, reason = conf_engine.should_confirm(task_context)

        assert should_confirm is True
        assert "risk" in reason.lower() or "critical" in reason.lower()

    def test_should_confirm_low_confidence(self, temp_policy_store):
        policy_store, _ = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T1",
            "task_description": "Update API schema",
            "command": "apply database schema update",
        }

        should_confirm, reason = conf_engine.should_confirm(task_context)

        assert should_confirm is True
        assert "insufficient" in reason.lower() or "first-time" in reason.lower() or "no confirmed rule" in reason.lower()

    def test_get_confidence_info(self, temp_policy_store):
        policy_store, project_scope = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T2",
            "task_description": "Run tests",
            "command": "python -m pytest tests/",
            "project_path": project_scope,
        }

        confidence_info = conf_engine.get_confidence_info(task_context)

        assert confidence_info["max_confidence"] >= 0.9
        assert confidence_info["count"] > 0

    def test_record_feedback_user_cancelled(self, temp_policy_store):
        policy_store, project_scope = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T2",
            "task_description": "Run tests",
            "command": "python -m pytest tests/",
            "project_path": project_scope,
        }

        decision = conf_engine.evaluate_task(task_context)
        conf_engine.record_feedback(
            task_context,
            was_confirmed=True,
            user_cancelled=True,
            decision_id=decision["decision_id"],
            execution_result="cancel",
            user_override="prefer_confirmation",
        )

        rules = policy_store.load_rules()
        rule = rules.get("rule_test_high_conf")

        assert rule is not None
        assert rule.last_seen_at

    def test_record_feedback_auto_executed(self, temp_policy_store):
        policy_store, project_scope = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T2",
            "task_description": "Run tests",
            "command": "python -m pytest tests/",
            "project_path": project_scope,
        }

        original_confidence = policy_store.get_rule("rule_test_high_conf").confidence
        decision = conf_engine.evaluate_task(task_context)

        conf_engine.record_feedback(
            task_context,
            was_confirmed=False,
            user_cancelled=False,
            decision_id=decision["decision_id"],
            execution_result="success",
        )

        rules = policy_store.load_rules()
        rule = rules.get("rule_test_high_conf")

        assert rule is not None
        assert rule.confidence >= original_confidence
        assert rule.last_applied_at

    def test_get_explanation_auto_execute(self, temp_policy_store):
        policy_store, project_scope = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T2",
            "task_description": "Run tests",
            "command": "python -m pytest tests/",
            "project_path": project_scope,
        }

        decision = conf_engine.evaluate_task(task_context)
        should_confirm = decision["final_decision"] == "require_confirmation"
        reason = decision["reason"]
        explanation = conf_engine.get_explanation(task_context, should_confirm, reason)

        assert "auto-execute" in explanation.lower()
        assert "project" in explanation.lower()
        assert "global" in explanation.lower()

    def test_get_explanation_need_confirm(self, temp_policy_store):
        policy_store, project_scope = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        task_context = {
            "task_type": "T3",
            "task_description": "Clean workspace directory",
            "command": "rm -rf workspace",
            "project_path": project_scope,
        }

        decision = conf_engine.evaluate_task(task_context)
        should_confirm = decision["final_decision"] == "require_confirmation"
        reason = decision["reason"]
        explanation = conf_engine.get_explanation(task_context, should_confirm, reason)

        assert "confirmation required" in explanation.lower()
        assert "git history" in explanation.lower() or "heuristic" in explanation.lower()

    def test_project_confirmed_rule_overrides_global_rule(self, temp_policy_store):
        policy_store, project_scope = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        decision = conf_engine.evaluate_task(
            {
                "task_type": "T2",
                "task_description": "Run tests",
                "command": "python -m pytest tests/",
                "project_path": project_scope,
            }
        )

        assert decision["final_decision"] == "auto_execute"
        assert decision["resolution"] == "project_over_global"
        assert len(decision["matched_rules"]) >= 2

    def test_first_time_low_risk_task_uses_heuristic_fallback(self):
        conf_engine = IntelligentConfirmation()
        decision = conf_engine.evaluate_task(
            {
                "task_type": "T1",
                "task_description": "Show test summary",
                "command": "pytest --collect-only -q",
            }
        )

        assert decision["final_decision"] == "auto_execute"
        assert "first-time" in decision["fallback_reason"]
        assert decision["heuristic_basis"]

    def test_high_risk_weak_hint_cannot_auto_execute(self, temp_policy_store):
        policy_store, project_scope = temp_policy_store
        conf_engine = IntelligentConfirmation(policy_store)
        decision = conf_engine.evaluate_task(
            {
                "task_type": "T3",
                "task_description": "Clean workspace directory",
                "command": "rm -rf workspace",
                "project_path": project_scope,
            }
        )

        assert decision["final_decision"] == "require_confirmation"
        assert any(rule["status"] == "hint" for rule in decision["matched_rules"])
        assert any("rm" in basis.lower() for basis in decision["heuristic_basis"])


class TestRiskAssessment:
    """Test risk assessment logic."""

    @pytest.fixture
    def conf_engine(self):
        return IntelligentConfirmation()

    def test_detect_critical_keywords(self, conf_engine):
        dangerous_tasks = [
            {"task_description": "Remove directory", "command": "rm -rf /data"},
            {"task_description": "Format disk", "command": "mkfs.ext4 /dev/sda1"},
            {"task_description": "Force push git", "command": "git push --force"},
        ]

        for task in dangerous_tasks:
            risk = conf_engine.assess_risk(task)
            assert risk == RiskLevel.CRITICAL

    def test_detect_medium_keywords(self, conf_engine):
        medium_tasks = [
            {"task_description": "Update database schema"},
            {"task_description": "Make API request"},
            {"task_description": "Install dependencies"},
        ]

        for task in medium_tasks:
            risk = conf_engine.assess_risk(task)
            assert risk == RiskLevel.MEDIUM

    def test_detect_low_keywords(self, conf_engine):
        low_tasks = [
            {"task_description": "Run tests"},
            {"task_description": "Check code format"},
            {"task_description": "List files"},
        ]

        for task in low_tasks:
            risk = conf_engine.assess_risk(task)
            assert risk == RiskLevel.LOW
